"""Provide a concrete implementation for UCLA dataset."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import List, Union

from ....api.decorators import register_datagrabber
from ....datagrabber import PatternDataGrabber
from ....utils import raise_error


@register_datagrabber
class JuselessUCLA(PatternDataGrabber):
    """Concrete implementation for pattern-based data fetching of UCLA data.

    Parameters
    ----------
    datadir : str or Path, optional
        The directory where the dataset is stored
        (default '/data/project/psychosis_thalamus/data/fmriprep').
    tasks : {"rest", "bart", "bht", "pamenc", "pamret", \
            "scap", "taskswitch", "stopsignal"} or \
            list of the options, optional
        UCLA task sessions. If None, all available task sessions are
        selected (default None).

    """

    def __init__(
        self,
        datadir: Union[
            str, Path
        ] = "/data/project/psychosis_thalamus/data/fmriprep",
        tasks: Union[str, List[str], None] = None,
    ) -> None:
        types = [
            "BOLD",
            "BOLD_confounds",
            "T1w",
            "probseg_CSF",
            "probseg_GM",
            "probseg_WM",
        ]

        if isinstance(tasks, str):
            tasks = [tasks]

        all_tasks = [
            "rest",
            "bart",
            "bht",
            "pamenc",
            "pamret",
            "scap",
            "taskswitch",
            "stopsignal",
        ]

        if tasks is None:
            tasks = all_tasks
        else:
            for t in tasks:
                if t not in all_tasks:
                    raise_error(
                        f"{t} is not a valid task in the UCLA dataset!"
                    )

        self.tasks = tasks

        patterns = {
            "BOLD": (
                "sub-{subject}/func/sub-{subject}_task-{task}_bold_space-"
                "MNI152NLin2009cAsym_preproc.nii.gz"
            ),
            "BOLD_confounds": (
                "sub-{subject}/func/sub-{subject}_"
                "task-{task}_bold_confounds.tsv"
            ),
            "T1w": (
                "sub-{subject}/anat/sub-{subject}_"
                "T1w_space-MNI152NLin2009cAsym_preproc.nii.gz"
            ),
            "probseg_CSF": (
                "sub-{subject}/anat/sub-{subject}_T1w_space-"
                "MNI152NLin2009cAsym_class-CSF_probtissue.nii.gz"
            ),
            "probseg_GM": (
                "sub-{subject}/anat/sub-{subject}_T1w_space-"
                "MNI152NLin2009cAsym_class-GM_probtissue.nii.gz"
            ),
            "probseg_WM": (
                "sub-{subject}/anat/sub-{subject}_T1w_space"
                "-MNI152NLin2009cAsym_class-WM_probtissue.nii.gz"
            ),
        }

        # the commented out uri leads to new open neuro dataset which does
        # NOT have preprocessed data
        # uri = "https://github.com/OpenNeuroDatasets/ds000030.git"

        replacements = ["subject", "task"]
        super().__init__(
            types=types,
            datadir=datadir,
            patterns=patterns,
            replacements=replacements,
        )

    def get_elements(self) -> List:
        """Implement fetching list of elements in the dataset.

        Returns
        -------
        elements : list
            The list of elements that can be grabbed in the dataset after
            imposing constraints based on specified tasks.
        """
        return [x for x in super().get_elements() if x[1] in self.tasks]



"""Provide concrete implementations for AOMICPIOP2 data access."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Vera Komeyer <v.komeyer@fz-juelich.de>
#          Xuan Li <xu.li@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import List, Union

from junifer.datagrabber import PatternDataladDataGrabber

from ...api.decorators import register_datagrabber
from ...utils import raise_error


@register_datagrabber
class DataladAOMICPIOP2(PatternDataladDataGrabber):
    """Concrete implementation for pattern-based data fetching of AOMICPIOP2.

    Parameters
    ----------
    datadir : str or Path, optional
        The directory where the datalad dataset will be cloned. If None,
        the datalad dataset will be cloned into a temporary directory
        (default None).
    tasks : {"restingstate", "stopsignal", "emomatching", "workingmemory"} \
            or list of the options, optional
        AOMIC PIOP2 task sessions. If None, all available task sessions are
        selected (default None).
    """

    def __init__(
        self,
        datadir: Union[str, Path, None] = None,
        tasks: Union[str, List[str], None] = None,
    ) -> None:
        # The types of data
        types = [
            "BOLD",
            "BOLD_confounds",
            "T1w",
            "probseg_CSF",
            "probseg_GM",
            "probseg_WM",
            "DWI",
        ]

        if isinstance(tasks, str):
            tasks = [tasks]

        all_tasks = [
            "restingstate",
            "emomatching",
            "workingmemory",
            "stopsignal",
        ]

        if tasks is None:
            tasks = all_tasks
        else:
            for t in tasks:
                if t not in all_tasks:
                    raise_error(
                        f"{t} is not a valid task in the AOMIC PIOP2"
                        " dataset!"
                    )

        self.tasks = tasks

        patterns = {
            "BOLD": (
                "derivatives/fmriprep/sub-{subject}/func/"
                "sub-{subject}_task-{task}_acq-seq_"
                "space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
            ),
            "BOLD_confounds": (
                "derivatives/fmriprep/sub-{subject}/func/"
                "sub-{subject}_task-{task}_acq-seq_"
                "desc-confounds_regressors.tsv"
            ),
            "T1w": (
                "derivatives/fmriprep/sub-{subject}/anat/"
                "sub-{subject}_space-MNI152NLin2009cAsym_"
                "desc-preproc_T1w.nii.gz"
            ),
            "probseg_CSF": (
                "derivatives/fmriprep/sub-{subject}/anat/"
                "sub-{subject}_space-MNI152NLin2009cAsym_label-"
                "CSF_probseg.nii.gz"
            ),
            "probseg_GM": (
                "derivatives/fmriprep/sub-{subject}/anat/"
                "sub-{subject}_space-MNI152NLin2009cAsym_label-"
                "GM_probseg.nii.gz"
            ),
            "probseg_WM": (
                "derivatives/fmriprep/sub-{subject}/anat/"
                "sub-{subject}_space-MNI152NLin2009cAsym_label-"
                "WM_probseg.nii.gz"
            ),
            "DWI": (
                "derivatives/dwipreproc/sub-{subject}/dwi/"
                "sub-{subject}_desc-preproc_dwi.nii.gz"
            ),
        }
        uri = "https://github.com/OpenNeuroDatasets/ds002790"
        replacements = ["subject", "task"]
        super().__init__(
            types=types,
            datadir=datadir,
            uri=uri,
            patterns=patterns,
            replacements=replacements,
        )

    def get_elements(self) -> List:
        """Implement fetching list of elements in the dataset.

        Returns
        -------
        list
            The list of elements that can be grabbed in the dataset after
            imposing constraints based on specified tasks.
        """
        all_elements = super().get_elements()
        return [x for x in all_elements if x[1] in self.tasks]



"""Provide utility functions for the storage sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import hashlib
import json
from importlib.metadata import PackageNotFoundError, version
from typing import Dict, Tuple

from ..utils.logging import logger, raise_error


def get_dependency_version(dependency: str) -> str:
    """Get dependency version.

    Parameters
    ----------
    dependency : str
         The dependency to fetch version for.

    Returns
    -------
    str
        The version of the dependency.

    """
    dep_version = ""
    try:
        dep_version = version(dependency)
    except PackageNotFoundError as e:
        raise_error(
            f"Could not obtain the version of {dependency}. "
            "Have you specified the DEPENDENCIES variable correctly?",
            exception=e,
        )

    return dep_version


def _meta_hash(meta: Dict) -> str:
    """Compute the MD5 hash of the metadata.

    Parameters
    ----------
    meta : dict
        The metadata as a dictionary. Must contain the key "element".

    Returns
    -------
    str
        The MD5 hash of the metadata.

    """
    logger.debug(f"Hashing metadata: {meta}")
    if "dependencies" not in meta:
        raise_error("The metadata must contain the key 'dependencies'")
    # Convert dependencies set into {dependency: version} dictionary
    meta["dependencies"] = {
        dep: get_dependency_version(dep) for dep in meta["dependencies"]
    }
    meta_md5 = hashlib.md5(
        json.dumps(meta, sort_keys=True).encode("utf-8")
    ).hexdigest()
    logger.debug(f"Hash computed: {meta_md5}")
    return meta_md5


def process_meta(meta: Dict) -> Tuple[str, Dict, Dict]:
    """Process the metadata for storage.

    It removes the key "element" and adds the "_element_keys" with the keys
    used to index the element.

    Parameters
    ----------
    meta : dict
        The metadata as a dictionary. Must contain the key "element".

    Returns
    -------
    str
        The MD5 hash of the metadata.
    dict
        The processed metadata for storage.
    tuple
        The element.

    Raises
    ------
    ValueError
        If `meta` is None or if it does not contain the key "element".

    """
    if meta is None:
        raise_error(msg="`meta` must be a dict (currently is None)")
    # Copy the metadata
    t_meta = meta.copy()
    # Remove key "element"
    element = t_meta.pop("element", None)
    if element is None:
        raise_error(msg="`meta` must contain the key 'element'")
    if "marker" not in t_meta:
        raise_error(msg="`meta` must contain the key 'marker'")
    if "name" not in t_meta["marker"]:
        raise_error(msg="`meta['marker']` must contain the key 'name'")
    if "type" not in t_meta:
        raise_error(msg="`meta` must contain the key 'type'")

    t_meta["_element_keys"] = list(element.keys())
    type_ = t_meta["type"]
    name = t_meta["marker"]["name"]
    t_meta["name"] = f"{type_}_{name}"
    # MD5 hash of the metadata
    md5_hash = _meta_hash(t_meta)
    return md5_hash, t_meta, element


def element_to_prefix(element: Dict) -> str:
    """Convert the element metadata to prefix.

    Parameters
    ----------
    element : dict
        The element to convert to prefix.

    Returns
    -------
    str
        The element converted to prefix.
    """
    logger.debug(f"Converting element {element} to prefix.")
    prefix = "element"
    if not isinstance(element, dict):
        raise_error(msg="`element` must be a dict")

    prefix = f"{prefix}_{'_'.join([f'{x}' for x in element.values()])}"

    logger.debug(f"Converted prefix: {prefix}")
    return f"{prefix}_"



"""Provide tests for datalad_base."""

# Authors: Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import Type

import datalad.api as dl
import pytest

from junifer.datagrabber.datalad_base import DataladDataGrabber


_testing_dataset = {
    "example_bids": {
        "uri": "https://gin.g-node.org/juaml/datalad-example-bids",
        "commit": "522dfb203afcd2cd55799bf347f9b211919a7338",
        "id": "fec92475-d9c0-4409-92ba-f041b6a12c40",
    },
    "example_bids_ses": {
        "uri": "https://gin.g-node.org/juaml/datalad-example-bids-ses",
        "commit": "3d08d55d1faad4f12ab64ac9497544a0d924d47a",
        "id": "c83500d0-532f-45be-baf1-0dab703bdc2a",
    },
}


def test_datalad_base_abstractness() -> None:
    """Test datalad base is abstract."""
    with pytest.raises(TypeError, match=r"abstract"):
        DataladDataGrabber()  # type: ignore


@pytest.fixture
def concrete_datagrabber() -> Type[DataladDataGrabber]:
    """Return a concrete datagrabber class.

    Returns
    -------
    DataladDataGrabber
        A concrete datagrabber class.

    """

    class MyDataGrabber(DataladDataGrabber):  # type: ignore
        def __init__(self, datadir, uri):
            super().__init__(
                datadir=datadir,
                rootdir="example_bids",
                uri=uri,
                types=["T1w", "BOLD"],
            )

        def get_item(self, subject):
            out = {
                "T1w": {
                    "path": self.datadir
                    / f"{subject}/anat/{subject}_T1w.nii.gz"
                },
                "BOLD": {
                    "path": self.datadir
                    / f"{subject}/func/{subject}_task-rest_bold.nii.gz"
                },
            }
            return out

        def get_elements(self):
            return [f"sub-{i:02d}" for i in range(1, 10)]

        def get_element_keys(self):
            return ["subject"]

    return MyDataGrabber


def test_datalad_install_errors(
    tmp_path: Path, concrete_datagrabber: Type
) -> None:
    """Test datalad base install errors / warnings.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.
    concrete_datagrabber : DataladDataGrabber
        A concrete datagrabber class to use.
    """

    # Dataset cloned outside of datagrabber
    datadir = tmp_path / "cloned_uri"
    uri = _testing_dataset["example_bids"]["uri"]
    uri2 = _testing_dataset["example_bids_ses"]["uri"]

    # Files are not there
    assert datadir.exists() is False
    # Clone dataset
    dl.clone(uri, datadir)  # type: ignore
    dg = concrete_datagrabber(datadir=datadir, uri=uri2)
    with pytest.raises(ValueError, match=r"different ID"):
        with dg:
            pass

    elem1_t1w = datadir / "example_bids/sub-01/anat/sub-01_T1w.nii.gz"
    elem1_t1w.unlink()
    with open(elem1_t1w, "w") as f:
        f.write("modified!")

    dg = concrete_datagrabber(datadir=datadir, uri=uri)
    with pytest.warns(RuntimeWarning, match=r"one file is not clean"):
        with dg:
            pass


def test_datalad_clone_cleanup(
    tmp_path: Path, concrete_datagrabber: Type
) -> None:
    """Test datalad base clone and remove.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.
    concrete_datagrabber : DataladDataGrabber
        A concrete datagrabber class to use.
    """

    # Clone whole dataset
    datadir = tmp_path / "newclone"
    uri = _testing_dataset["example_bids"]["uri"]
    elem1_bold = (
        datadir / "example_bids/sub-01/func/sub-01_task-rest_bold.nii.gz"
    )
    elem1_t1w = datadir / "example_bids/sub-01/anat/sub-01_T1w.nii.gz"

    assert datadir.exists() is False
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_file() is False
    with concrete_datagrabber(datadir=datadir, uri=uri) as dg:
        assert datadir.exists() is True
        assert dg._was_cloned is True
        assert elem1_bold.is_file() is False
        assert elem1_bold.is_symlink() is True
        assert elem1_t1w.is_file() is False
        assert elem1_t1w.is_symlink() is True
        elem1 = dg["sub-01"]
        assert "meta" in elem1["BOLD"]
        meta = elem1["BOLD"]["meta"]
        assert "datagrabber" in meta
        assert "datalad_dirty" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_dirty"] is False
        assert hasattr(dg, "_got_files") is False
        assert datadir.exists() is True
        assert elem1_bold.is_file() is True
        assert elem1_bold.is_symlink() is True
        assert elem1_t1w.is_file() is True
        assert elem1_t1w.is_symlink() is True

    assert datadir.exists() is False
    assert len(list(datadir.glob("*"))) == 0


def test_datalad_previously_cloned(
    tmp_path: Path, concrete_datagrabber: Type
) -> None:
    """Test datalad base on cloned dataset.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.
    concrete_datagrabber : DataladDataGrabber
        A concrete datagrabber class to use.
    """

    # Dataset cloned outside of datagrabber
    datadir = tmp_path / "cloned"
    elem1_bold = (
        datadir / "example_bids/sub-01/func/sub-01_task-rest_bold.nii.gz"
    )
    elem1_t1w = datadir / "example_bids/sub-01/anat/sub-01_T1w.nii.gz"
    uri = _testing_dataset["example_bids"]["uri"]
    commit = _testing_dataset["example_bids"]["commit"]
    remote_id = _testing_dataset["example_bids"]["id"]
    # Files are not there
    assert datadir.exists() is False
    assert elem1_bold.exists() is False
    assert elem1_t1w.exists() is False

    # Clone dataset
    dl.clone(uri, datadir, result_renderer="disabled")  # type: ignore

    # Files are there, but are empty symbolic links
    assert datadir.exists() is True
    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is True
    assert elem1_t1w.is_file() is False
    with concrete_datagrabber(datadir=datadir, uri=uri) as dg:
        assert datadir.exists() is True
        assert dg._was_cloned is False
        elem1 = dg["sub-01"]
        assert "meta" in elem1["BOLD"]
        meta = elem1["BOLD"]["meta"]
        assert "datagrabber" in meta
        assert "datalad_dirty" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_dirty"] is False
        assert "datalad_commit_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_commit_id"] == commit
        assert "datalad_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_id"] == remote_id

        assert hasattr(dg, "_got_files") is True
        # Files are there and symlinks are fixed
        assert elem1["BOLD"]["path"].is_file() is True
        assert elem1["BOLD"]["path"].is_symlink() is True
        assert elem1["T1w"]["path"].is_file() is True
        assert elem1["T1w"]["path"].is_symlink() is True

        # Datagrabber fetched two files
        assert len(dg._got_files) == 2
        assert any(x.name == "sub-01_T1w.nii.gz" for x in dg._got_files)
        assert any(
            x.name == "sub-01_task-rest_bold.nii.gz" for x in dg._got_files
        )

    assert datadir.exists() is True
    assert len(list(datadir.glob("*"))) > 0


def test_datalad_previously_cloned_and_get(
    tmp_path: Path, concrete_datagrabber: Type
) -> None:
    """Test datalad base on cloned dataset with files present.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.
    concrete_datagrabber : DataladDataGrabber
        A concrete datagrabber class to use.
    """

    # Dataset cloned outside of datagrabber with some files present
    datadir = tmp_path / "cloned_clean"
    elem1_bold = (
        datadir / "example_bids/sub-01/func/sub-01_task-rest_bold.nii.gz"
    )
    elem1_t1w = datadir / "example_bids/sub-01/anat/sub-01_T1w.nii.gz"
    uri = _testing_dataset["example_bids"]["uri"]
    commit = _testing_dataset["example_bids"]["commit"]
    remote_id = _testing_dataset["example_bids"]["id"]

    # Files are not there
    assert datadir.exists() is False
    assert elem1_bold.exists() is False
    assert elem1_t1w.exists() is False

    # Clone dataset
    dl.clone(uri, datadir, result_renderer="disabled")  # type: ignore

    # Files are there, but are empty symbolic links
    assert datadir.exists() is True
    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is True
    assert elem1_t1w.is_file() is False

    dl.get(  # type: ignore
        elem1_t1w, dataset=datadir, result_renderer="disabled"
    )

    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is True
    assert elem1_t1w.is_file() is True

    with concrete_datagrabber(datadir=datadir, uri=uri) as dg:
        assert datadir.exists() is True
        assert dg._was_cloned is False
        elem1 = dg["sub-01"]
        assert "meta" in elem1["BOLD"]
        meta = elem1["BOLD"]["meta"]
        assert "datagrabber" in meta
        assert "datalad_dirty" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_dirty"] is False
        assert "datalad_commit_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_commit_id"] == commit
        assert "datalad_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_id"] == remote_id

        assert hasattr(dg, "_got_files") is True
        # Files are there and symlinks are fixed
        assert elem1["BOLD"]["path"].is_file() is True
        assert elem1["BOLD"]["path"].is_symlink() is True
        assert elem1["T1w"]["path"].is_file() is True
        assert elem1["T1w"]["path"].is_symlink() is True

        # Datagrabber fetched two files
        assert len(dg._got_files) == 1
        assert dg._got_files[0].name == "sub-01_task-rest_bold.nii.gz"

    assert datadir.exists() is True
    assert len(list(datadir.glob("*"))) > 0

    # Same state as before using the datagrabber
    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is True
    assert elem1_t1w.is_file() is True


def test_datalad_previously_cloned_and_get_dirty(
    tmp_path: Path, concrete_datagrabber: Type
) -> None:
    """Test datalad base on a dirty cloned dataset.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.
    concrete_datagrabber : DataladDataGrabber
        A concrete datagrabber class to use.
    """

    # Dataset cloned outside of datagrabber with some files present and dirty
    datadir = tmp_path / "cloned_dirty"
    elem1_bold = (
        datadir / "example_bids/sub-01/func/sub-01_task-rest_bold.nii.gz"
    )
    elem1_t1w = datadir / "example_bids/sub-01/anat/sub-01_T1w.nii.gz"
    uri = _testing_dataset["example_bids"]["uri"]
    commit = _testing_dataset["example_bids"]["commit"]
    remote_id = _testing_dataset["example_bids"]["id"]

    # Files are not there
    assert datadir.exists() is False
    assert elem1_bold.exists() is False
    assert elem1_t1w.exists() is False

    # Clone dataset
    dl.clone(uri, datadir, result_renderer="disabled")  # type: ignore

    # Files are there, but are empty symbolic links
    assert datadir.exists() is True
    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is True
    assert elem1_t1w.is_file() is False

    dl.get(  # type: ignore
        elem1_t1w, dataset=datadir, result_renderer="disabled"
    )

    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is True
    assert elem1_t1w.is_file() is True

    elem1_t1w.unlink()
    with open(elem1_t1w, "w") as f:
        f.write("modified!")

    with concrete_datagrabber(datadir=datadir, uri=uri) as dg:
        assert datadir.exists() is True
        assert dg._was_cloned is False
        elem1 = dg["sub-01"]
        assert "meta" in elem1["BOLD"]
        meta = elem1["BOLD"]["meta"]
        assert "datagrabber" in meta
        assert "datalad_dirty" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_dirty"] is True
        assert "datalad_commit_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_commit_id"] == commit
        assert "datalad_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_id"] == remote_id

        assert hasattr(dg, "_got_files") is True
        # Files are there and symlinks are fixed
        assert elem1["BOLD"]["path"].is_file() is True
        assert elem1["BOLD"]["path"].is_symlink() is True
        assert elem1["T1w"]["path"].is_file() is True
        assert elem1["T1w"]["path"].is_symlink() is False

        # Datagrabber fetched two files
        assert len(dg._got_files) == 1
        assert dg._got_files[0].name == "sub-01_task-rest_bold.nii.gz"

    # Now get another subject that has not been modified
    with concrete_datagrabber(datadir=datadir, uri=uri) as dg:
        assert datadir.exists() is True
        assert dg._was_cloned is False
        elem2 = dg["sub-02"]
        assert "meta" in elem1["BOLD"]
        meta = elem2["BOLD"]["meta"]
        assert "datagrabber" in meta
        assert "datalad_dirty" in meta["datagrabber"]

        # Dataset is still dirty due to subject sub-01
        assert meta["datagrabber"]["datalad_dirty"] is True

        assert "datalad_commit_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_commit_id"] == commit
        assert "datalad_id" in meta["datagrabber"]
        assert meta["datagrabber"]["datalad_id"] == remote_id

        assert hasattr(dg, "_got_files") is True
        # Files are there and symlinks are fixed
        assert elem2["BOLD"]["path"].is_file() is True
        assert elem2["BOLD"]["path"].is_symlink() is True
        assert elem2["T1w"]["path"].is_file() is True
        assert elem2["T1w"]["path"].is_symlink() is True

        # Datagrabber fetched two files
        assert len(dg._got_files) == 2
        assert any(x.name == "sub-02_T1w.nii.gz" for x in dg._got_files)
        assert any(
            x.name == "sub-02_task-rest_bold.nii.gz" for x in dg._got_files
        )

    assert datadir.exists() is True
    assert len(list(datadir.glob("*"))) > 0

    # Same state as before using the datagrabber
    assert elem1_bold.is_symlink() is True
    assert elem1_bold.is_file() is False
    assert elem1_t1w.is_symlink() is False
    assert elem1_t1w.is_file() is True



"""Provide utility functions shared by different markers."""

# Authors: Leonard Sasse <l.sasse@fz-juelich.de>
#          Nicolás Nieto <n.nieto@fz-juelich.de>
#          Sami Hamdan <s.hamdan@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Callable, Union

import numpy as np
import pandas as pd
from scipy.stats import zscore

from ..utils import raise_error


def _ets(bold_ts: np.ndarray) -> np.ndarray:
    """Compute the edge-wise time series based on BOLD time series.

    Take a timeseries of brain areas, and calculate timeseries for each
    edge according to the method outlined in [1]_. For more information,
    check https://github.com/brain-networks/edge-ts/blob/master/main.m

    Parameters
    ----------
    bold_ts : np.ndarray
        BOLD time series (time x ROIs)

    Returns
    -------
    np.ndarray
        edge-wise time series, i.e. estimate of functional connectivity at each
        time point.

    References
    ----------
    .. [1] Zamani Esfahlani et al. (2020)
            High-amplitude cofluctuations in cortical activity drive
            functional connectivity
            doi: 10.1073/pnas.2005531117

    """
    # Compute the z-score for each brain region's timeseries
    timeseries = zscore(bold_ts)
    # Get the number of ROIs
    _, n_roi = timeseries.shape
    # indices of unique edges (lower triangle)
    u, v = np.tril_indices(n_roi, k=-1)
    # Compute the ETS
    return timeseries[:, u] * timeseries[:, v]


def _correlate_dataframes(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    method: Union[str, Callable] = "pearson",
) -> pd.DataFrame:
    """Column-wise correlations between two dataframes.

    Correlates each column of `df1` with each column of `df2`.
    Output is a dataframe of shape (df2.shape[1], df1.shape[1]).
    It is required that number of rows are matched.

    Parameters
    ----------
    df1 : pandas.DataFrame
        The first dataframe.
    df2 : pandas.DataFrame
        The second dataframe.
    method : str or callable, optional
        any method that can be passed to
        :func:`pandas.DataFrame.corr` (default "pearson").

    Returns
    -------
    df_corr : pandas.DataFrame
        The correlated values as a dataframe.

    Raises
    ------
    ValueError
        If number of rows between dataframes are not matched.

    """

    if df1.shape[0] != df2.shape[0]:
        raise_error("pandas.DataFrame's have unequal number of rows!")
    return (
        pd.concat([df1, df2], axis=1, keys=["df1", "df2"])  # type: ignore
        .corr(method=method)  # type: ignore
        .loc["df2", "df1"]
    )



"""Provide mixin class for updating metadata."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Dict


class UpdateMetaMixin:
    """Mixin class for updating meta."""

    def update_meta(
        self,
        input: Dict,
        step_name: str,
    ) -> None:
        """Update metadata.

        Parameters
        ----------
        input : dict
            The data object to update.
        step_name : str
            The name of the pipeline step.
        """
        t_meta = {}
        t_meta["class"] = self.__class__.__name__
        for k, v in vars(self).items():
            if not k.startswith("_"):
                t_meta[k] = v

        if "meta" not in input:
            input["meta"] = {}
        input["meta"][step_name] = t_meta
        if "dependencies" not in input["meta"]:
            input["meta"]["dependencies"] = set()

        dependencies = getattr(self, "_DEPENDENCIES", set())
        if dependencies is not None:
            if not isinstance(dependencies, (set, list)):
                dependencies = set([dependencies])
            input["meta"]["dependencies"].update(dependencies)



# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL
from tempfile import TemporaryDirectory
from pathlib import Path

import datalad.api as dl

dst = 'git@gin.g-node.org:/juaml/datalad-example-bids.git'

with TemporaryDirectory() as tmpdir_name:
    tmpdir = Path(tmpdir_name)
    ds = dl.create(tmpdir)  # type: ignore

    base_dir = tmpdir / 'example_bids'
    base_dir.mkdir()

    for i_sub in range(1, 10):
        t_sub = f'sub-{i_sub:02d}'
        sub_dir = base_dir / t_sub
        sub_dir.mkdir()

        for dname in ['anat', 'func']:
            (sub_dir / dname).mkdir()

        fnames = [f'anat/{t_sub}_T1w.nii.gz',
                  f'func/{t_sub}_task-rest_bold.nii.gz',
                  f'func/{t_sub}_task-rest_bold.json']
        for fname in fnames:
            with open(sub_dir / fname, 'w') as f:
                f.write(f'placeholder-{fname}')

    ds.save(recursive=True)
    ds.siblings('add', name='gin', url=dst)
    ds.push(to='gin', force='all')



"""Provide tests for SPM Auditory datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from junifer.testing.datagrabbers import SPMAuditoryTestingDatagrabber


def test_SPMAuditoryTestingDatagrabber() -> None:
    """Test SPM Auditory datagrabber."""
    expected_elements = [
        "sub001",
        "sub002",
        "sub003",
        "sub004",
        "sub005",
        "sub006",
        "sub007",
        "sub008",
        "sub009",
        "sub010",
    ]
    with SPMAuditoryTestingDatagrabber() as dg:
        all_elements = dg.get_elements()
        assert set(all_elements) == set(expected_elements)
        out = dg["sub001"]
        assert "BOLD" in out
        assert out["BOLD"]["path"].exists()
        assert out["BOLD"]["path"].is_file()

        assert "T1w" in out
        assert out["T1w"]["path"].exists()
        assert out["T1w"]["path"].is_file()



"""Provide abstract base class for multiple source datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Dict, List, Tuple, Union

from .base import BaseDataGrabber


class MultipleDataGrabber(BaseDataGrabber):
    """Data Grabber class for data fetching from multiple sources.

    Defines a Data Grabber which can be used to fetch data from multiple
    datagrabbers.

    Parameters
    ----------
    datagrabbers : list of datagrabber-like objects
        The datagrabbers to use to fetch data using.
    **kwargs
        Keyword arguments passed to superclass.
    """

    def __init__(self, datagrabbers: List[BaseDataGrabber], **kwargs) -> None:
        # Check datagrabbers consistency
        # 1) same element keys
        first_keys = datagrabbers[0].get_element_keys()
        for dg in datagrabbers[1:]:
            if dg.get_element_keys() != first_keys:
                raise ValueError("Datagrabbers have different element keys.")
        # 2) no overlapping types
        types = [x for dg in datagrabbers for x in dg.get_types()]
        if len(types) != len(set(types)):
            raise ValueError("Datagrabbers have overlapping types.")
        self._datagrabbers = datagrabbers

    def __getitem__(self, element: Union[str, Tuple]) -> Dict:
        """Implement indexing.

        Parameters
        ----------
        element : str or tuple
            The element to be indexed. If one string is provided, it is
            assumed to be a tuple with only one item. If a tuple is provided,
            each item in the tuple is the value for the replacement string
            specified in "replacements".

        Returns
        -------
        dict
            Dictionary of paths for each type of data required for the
            specified element.

        """

        out = {}
        metas = []
        for dg in self._datagrabbers:
            t_out = dg[element]
            out.update(t_out)
            # Now get the meta for this datagrabber
            t_meta = {}
            dg.update_meta(t_meta, "datagrabber")
            # Store all the sub-datagrabbers meta
            metas.append(t_meta["meta"]["datagrabber"])

        # Update all the metas again
        for kind in out:
            self.update_meta(out[kind], "datagrabber")
            out[kind]["meta"]["datagrabber"]["datagrabbers"] = metas
        return out

    def get_item(self, **element: Dict) -> Dict[str, Dict]:
        """Get item.

        Parameters
        ----------
        element : dict
            The element to be indexed.

        Returns
        -------
        dict
            Dictionary of paths for each type of data required for the
            specified element.

        Notes
        -----
            This function is not implemented for this class as it is useless.
        """
        raise NotImplementedError(
            "get_item() is not useful for this class, hence not implemented."
        )

    def __enter__(self) -> "BaseDataGrabber":
        """Implement context entry."""
        for dg in self._datagrabbers:
            dg.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        """Implement context exit."""
        for dg in self._datagrabbers:
            dg.__exit__(exc_type, exc_value, exc_traceback)

    def get_elements(self) -> List:
        """Get elements.

        Returns
        -------
        list
            The list of elements that can be grabbed in the dataset. It
            corresponds to the elements that are present in all the
            related datagrabbers.
        """
        all_elements = [dg.get_elements() for dg in self._datagrabbers]
        elements = set(all_elements[0])
        for s in all_elements[1:]:
            elements.intersection_update(s)
        return list(elements)

    def get_element_keys(self) -> List[str]:
        """Get element keys.

        For each item in the ``element`` tuple passed to ``__getitem__()``,
        this method returns the corresponding key(s).

        Returns
        -------
        list of str
            The element keys.
        """
        return self._datagrabbers[0].get_element_keys()

    def get_types(self) -> List[str]:
        """Get types.

        Returns
        -------
        list of list of str
            The types of data to be grabbed.

        """
        types = [x for dg in self._datagrabbers for x in dg.get_types()]
        return types



"""Provide imports for storage sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from .base import BaseFeatureStorage
from .pandas_base import PandasBaseFeatureStorage
from .sqlite import SQLiteFeatureStorage



"""Provide class for CamCAN VBM juseless datalad datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import Union

from ....api.decorators import register_datagrabber
from ....datagrabber import PatternDataladDataGrabber


@register_datagrabber
class JuselessDataladCamCANVBM(PatternDataladDataGrabber):
    """Juseless CamCAN VBM Data Grabber class.

    Implements a Data Grabber to access the CamCAN VBM data in Juseless.

    Parameters
    ----------
    datadir : str or pathlib.Path, optional
        The directory where the datalad dataset will be cloned. If None,
        the datalad dataset will be cloned into a temporary directory
        (default None).

    """

    def __init__(self, datadir: Union[str, Path, None] = None) -> None:
        uri = (
            "ria+http://cat_12.5.ds.inm7.de"
            "#a139b26a-8406-11ea-8f94-a0369f287950"
        )
        types = ["VBM_GM"]
        replacements = ["subject"]
        patterns = {"VBM_GM": "sub-{subject}/mri/m0wp1sub-{subject}.nii.gz"}
        super().__init__(
            types=types,
            datadir=datadir,
            uri=uri,
            replacements=replacements,
            patterns=patterns,
        )



"""Provide tests for multiple."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import pytest

from junifer.datagrabber import MultipleDataGrabber, PatternDataladDataGrabber


_testing_dataset = {
    "example_bids": {
        "uri": "https://gin.g-node.org/juaml/datalad-example-bids",
        "id": "522dfb203afcd2cd55799bf347f9b211919a7338",
    },
    "example_bids_ses": {
        "uri": "https://gin.g-node.org/juaml/datalad-example-bids-ses",
        "id": "3d08d55d1faad4f12ab64ac9497544a0d924d47a",
    },
}


def test_multiple() -> None:
    """Test a multiple datagrabber."""
    repo_uri = _testing_dataset["example_bids_ses"]["uri"]
    rootdir = "example_bids_ses"
    replacements = ["subject", "session"]
    pattern1 = {
        "T1w": "{subject}/{session}/anat/{subject}_{session}_T1w.nii.gz",
    }
    pattern2 = {
        "BOLD": "{subject}/{session}/func/"
        "{subject}_{session}_task-rest_bold.nii.gz",
    }
    dg1 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri,
        types=["T1w"],
        patterns=pattern1,
        replacements=replacements,
    )

    dg2 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri,
        types=["BOLD"],
        patterns=pattern2,
        replacements=replacements,
    )

    dg = MultipleDataGrabber([dg1, dg2])

    types = dg.get_types()
    assert "T1w" in types
    assert "BOLD" in types

    expected_subs = [
        (f"sub-{i:02d}", f"ses-{j:02d}")
        for j in range(1, 3)
        for i in range(1, 10)
    ]

    with dg:
        subs = [x for x in dg]
        assert set(subs) == set(expected_subs)

        elem = dg[("sub-01", "ses-01")]
        assert "T1w" in elem
        assert "BOLD" in elem
        assert "meta" in elem["BOLD"]
        meta = elem["BOLD"]["meta"]["datagrabber"]
        assert "class" in meta
        assert meta["class"] == "MultipleDataGrabber"
        assert "datagrabbers" in meta
        assert len(meta["datagrabbers"]) == 2
        assert meta["datagrabbers"][0]["class"] == "PatternDataladDataGrabber"
        assert meta["datagrabbers"][1]["class"] == "PatternDataladDataGrabber"


def test_multiple_no_intersection() -> None:
    """Test a multiple datagrabber without intersection (0 elements)."""
    repo_uri1 = _testing_dataset["example_bids"]["uri"]
    repo_uri2 = _testing_dataset["example_bids_ses"]["uri"]
    rootdir = "example_bids_ses"
    replacements = ["subject", "session"]
    pattern1 = {
        "T1w": "{subject}/{session}/anat/{subject}_{session}_T1w.nii.gz",
    }
    pattern2 = {
        "BOLD": "{subject}/{session}/func/"
        "{subject}_{session}_task-rest_bold.nii.gz",
    }
    dg1 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri1,
        types=["T1w"],
        patterns=pattern1,
        replacements=replacements,
    )

    dg2 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri2,
        types=["BOLD"],
        patterns=pattern2,
        replacements=replacements,
    )

    dg = MultipleDataGrabber([dg1, dg2])
    expected_subs = set()
    with dg:
        subs = [x for x in dg]
        assert set(subs) == set(expected_subs)


def test_multiple_get_item() -> None:
    """Test a multiple datagrabber get_item error."""
    repo_uri1 = _testing_dataset["example_bids"]["uri"]
    rootdir = "example_bids_ses"
    replacements = ["subject", "session"]
    pattern1 = {
        "T1w": "{subject}/{session}/anat/{subject}_{session}_T1w.nii.gz",
    }
    dg1 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri1,
        types=["T1w"],
        patterns=pattern1,
        replacements=replacements,
    )

    dg = MultipleDataGrabber([dg1])
    with pytest.raises(NotImplementedError):
        dg.get_item(subject="sub-01")  # type: ignore


def test_multiple_validation() -> None:
    """Test a multiple datagrabber init validation."""
    repo_uri1 = _testing_dataset["example_bids"]["uri"]
    repo_uri2 = _testing_dataset["example_bids_ses"]["uri"]
    rootdir = "example_bids_ses"
    replacement1 = ["subject", "session"]
    replacement2 = ["subject"]
    pattern1 = {
        "T1w": "{subject}/{session}/anat/{subject}_{session}_T1w.nii.gz",
    }
    pattern2 = {
        "bold": "{subject}/func/{subject}_task-rest_bold.nii.gz",
    }
    dg1 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri1,
        types=["T1w"],
        patterns=pattern1,
        replacements=replacement1,
    )

    dg2 = PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri2,
        types=["bold"],
        patterns=pattern2,
        replacements=replacement2,
    )

    with pytest.raises(ValueError, match="different element key"):
        MultipleDataGrabber([dg1, dg2])

    with pytest.raises(ValueError, match="overlapping types"):
        MultipleDataGrabber([dg1, dg1])



"""Provide utility functions for the datagrabber sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Dict, List

from ..utils import raise_error


def validate_types(types: List[str]) -> None:
    """Validate the types.

    Parameters
    ----------
    types : list of str
        The object to validate.

    """
    if not isinstance(types, list):
        raise_error(msg="`types` must be a list", klass=TypeError)
    if any(not isinstance(x, str) for x in types):
        raise_error(msg="`types` must be a list of strings", klass=TypeError)


def validate_replacements(
    replacements: List[str], patterns: Dict[str, str]
) -> None:
    """Validate the replacements.

    Parameters
    ----------
    replacements : list of str
        The object to validate.
    patterns : dict
        The patterns to validate against.

    """
    if not isinstance(replacements, list):
        raise_error(msg="`replacements` must be a list.", klass=TypeError)

    if not isinstance(patterns, dict):
        raise_error(msg="`patterns` must be a dict.", klass=TypeError)

    if any(not isinstance(x, str) for x in replacements):
        raise_error(
            msg="`replacements` must be a list of strings.", klass=TypeError
        )

    for x in replacements:
        if all(x not in y for y in patterns.values()):
            raise_error(msg=f"Replacement {x} is not part of any pattern.")

    # Check that at least one pattern has all the replacements
    at_least_one = False
    for _, v in patterns.items():
        if all(x in v for x in replacements):
            at_least_one = True
    if at_least_one is False:
        raise_error(msg="At least one pattern must contain all replacements.")


def validate_patterns(types: List[str], patterns: Dict[str, str]) -> None:
    """Validate the patterns.

    Parameters
    ----------
    types : list of str
        The types list.
    patterns : dict
        The object to validate.

    """
    # Validate the types
    validate_types(types)
    if not isinstance(patterns, dict):
        raise_error(msg="`patterns` must be a dict.", klass=TypeError)
    # Unequal length of objects
    if len(types) != len(patterns):
        raise_error(
            msg="`types` and `patterns` must have the same length.",
            klass=ValueError,
        )

    if any(x not in patterns for x in types):
        raise_error(
            msg="`patterns` must contain all `types`", klass=ValueError
        )

    if any("}*" in pattern for pattern in patterns.values()):
        raise_error(
            msg="`patterns` must not contain `*` following a replacement",
            klass=ValueError,
        )



"""Provide tests for pattern_datalad."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path

import pytest

from junifer.datagrabber.pattern_datalad import PatternDataladDataGrabber


_testing_dataset = {
    "example_bids": {
        "uri": "https://gin.g-node.org/juaml/datalad-example-bids",
        "commit": "522dfb203afcd2cd55799bf347f9b211919a7338",
        "id": "fec92475-d9c0-4409-92ba-f041b6a12c40",
    },
    "example_bids_ses": {
        "uri": "https://gin.g-node.org/juaml/datalad-example-bids-ses",
        "commit": "3d08d55d1faad4f12ab64ac9497544a0d924d47a",
        "id": "c83500d0-532f-45be-baf1-0dab703bdc2a",
    },
}


def test_bids_pattern_datalad_datagrabber_missing_uri() -> None:
    """Test check of missing URI in pattern datalad datagrabber."""
    with pytest.raises(ValueError, match=r"`uri` must be provided"):
        PatternDataladDataGrabber(
            datadir=None,
            types=[],
            patterns={},
            replacements=[],
        )


def test_bids_PatternDataladDataGrabber(tmp_path: Path) -> None:
    """Test a subject-based BIDS datalad datagrabber.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    # Define types
    types = ["T1w", "BOLD"]
    # Define patterns
    patterns = {
        "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
        "BOLD": "{subject}/func/{subject}_task-rest_bold.nii.gz",
    }
    # Define replacements
    replacements = ["subject"]

    repo_uri = _testing_dataset["example_bids"]["uri"]
    rootdir = "example_bids"
    repo_commit = _testing_dataset["example_bids"]["commit"]

    with PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri,
        types=types,
        patterns=patterns,
        replacements=replacements,
    ) as dg:
        subs = [x for x in dg]
        expected_subs = [f"sub-{i:02d}" for i in range(1, 10)]
        assert set(subs) == set(expected_subs)

        for elem in dg:
            t_sub = dg[elem]
            assert "path" in t_sub["T1w"]
            assert t_sub["T1w"]["path"] == (
                dg.datadir / f"{elem}/anat/{elem}_T1w.nii.gz"
            )
            assert "path" in t_sub["BOLD"]
            assert t_sub["BOLD"]["path"] == (
                dg.datadir / f"{elem}/func/{elem}_task-rest_bold.nii.gz"
            )

            assert "meta" in t_sub["BOLD"]
            meta = t_sub["BOLD"]["meta"]
            assert "datagrabber" in meta
            dg_meta = meta["datagrabber"]
            assert "class" in dg_meta
            assert dg_meta["class"] == "PatternDataladDataGrabber"
            assert "uri" in dg_meta
            assert dg_meta["uri"] == repo_uri
            assert "datalad_commit_id" in dg_meta
            assert dg_meta["datalad_commit_id"] == repo_commit

            with open(t_sub["T1w"]["path"], "r") as f:
                assert f.readlines()[0].startswith("placeholder")


def test_bids_PatternDataladDataGrabber_datadir(tmp_path: Path) -> None:
    """Test a datalad datagrabber with a datadir set to a relative path.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    # Define types
    types = ["T1w", "BOLD"]
    # Define patterns
    patterns = {
        "T1w": "{subject}/anat/{subject}_T1w.nii.gz",
        "BOLD": "{subject}/func/{subject}_task-rest_bold.nii.gz",
    }
    # Define replacements
    replacements = ["subject"]

    repo_uri = _testing_dataset["example_bids"]["uri"]

    datadir = "dataset"  # use string and not absolute path
    patterns = {
        "T1w": "example_bids/{subject}/anat/{subject}_T*w.nii.gz",
        "BOLD": "example_bids/{subject}/func/{subject}_task-rest_*.nii.gz",
    }
    with PatternDataladDataGrabber(
        uri=repo_uri,
        types=types,
        patterns=patterns,
        datadir=datadir,
        replacements=replacements,
    ) as dg:
        assert dg.datadir == Path(datadir)
        for elem in dg:
            t_sub = dg[elem]
            assert "path" in t_sub["T1w"]
            assert t_sub["T1w"]["path"] == (
                dg.datadir / f"{elem}/anat/{elem}_T1w.nii.gz"
            )
            assert "path" in t_sub["BOLD"]
            assert t_sub["BOLD"]["path"] == (
                dg.datadir / f"{elem}/func/{elem}_task-rest_bold.nii.gz"
            )


def test_bids_PatternDataladDataGrabber_session():
    """Test a subject and session-based BIDS datalad datagrabber."""
    types = ["T1w", "BOLD"]
    patterns = {
        "T1w": "{subject}/{session}/anat/{subject}_{session}_T1w.nii.gz",
        "BOLD": "{subject}/{session}/func/"
        "{subject}_{session}_task-rest_bold.nii.gz",
    }
    replacements = ["subject", "session"]

    with pytest.raises(ValueError, match=r"`uri` must be provided"):
        PatternDataladDataGrabber(
            datadir=None,
            types=types,
            patterns=patterns,
            replacements=replacements,
        )

    repo_uri = _testing_dataset["example_bids_ses"]["uri"]
    rootdir = "example_bids_ses"
    # repo_commit =  _testing_dataset['example_bids_ses']['id']

    # With T1W and bold, only 2 sessions are available
    with PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri,
        types=types,
        patterns=patterns,
        replacements=replacements,
    ) as dg:
        subs = [x for x in dg]
        expected_subs = [
            (f"sub-{i:02d}", f"ses-{j:02d}")
            for j in range(1, 3)
            for i in range(1, 10)
        ]
        assert set(subs) == set(expected_subs)

    # Test with a different T1w only, it should have 3 sessions
    types = ["T1w"]
    patterns = {
        "T1w": "{subject}/{session}/anat/{subject}_{session}_T1w.nii.gz",
    }
    with PatternDataladDataGrabber(
        rootdir=rootdir,
        uri=repo_uri,
        types=types,
        patterns=patterns,
        replacements=replacements,
    ) as dg:
        subs = [x for x in dg]
        expected_subs = [
            (f"sub-{i:02d}", f"ses-{j:02d}")
            for j in range(1, 4)
            for i in range(1, 10)
        ]
        assert set(subs) == set(expected_subs)



"""Provide abstract base class for datalad datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import tempfile
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import datalad.api as dl
from datalad.support.gitrepo import GitRepo

from ..api.decorators import register_datagrabber
from ..utils import logger, raise_error, warn_with_log
from .base import BaseDataGrabber


@register_datagrabber
class DataladDataGrabber(BaseDataGrabber):
    """Abstract base class for data fetching via Datalad.

    Defines a Data Grabber that gets data from a datalad sibling.

    Parameters
    ----------
    rootdir : str or Path, optional
        The path within the datalad dataset to the root directory
        (default ".").
    datadir : str or Path, optional
        That directory where the datalad dataset will be cloned. If None,
        the datalad dataset will be cloned into a temporary directory
        (default None).
    uri : str, optional
        URI of the datalad sibling (default None).
    **kwargs
        Keyword arguments passed to superclass.

    Methods
    -------
    install:
        Installs (clones) the datalad dataset into the `datadir`. This method
        is called automatically when the datagrabber is used within a context.
    remove:
        Removes the datalad dataset from the `datadir`. This method is called
        automatically when the datagrabber is used within a context.

    See Also
    --------
    BaseDataGrabber
        For the abstract base class of any datagrabber.

    Notes
    -----
    By itself, this class is still abstract as the `__getitem__` method relies
    on the parent class `BaseDataGrabber.__getitem__` which is not yet
    implemented. This class is intended to be used as a superclass of a class
    with multiple inheritance.
    """

    def __init__(
        self,
        rootdir: Union[str, Path] = ".",
        datadir: Union[str, Path, None] = None,
        uri: Optional[str] = None,
        **kwargs,
    ):
        if datadir is None:
            logger.warning("`datadir` is None, creating a temporary directory")
            # Create temporary directory
            datadir = tempfile.mkdtemp()
            logger.info(f"`datadir` set to {datadir}")
        # TODO: uri can be converted to a positional argument
        if uri is None:
            raise_error("`uri` must be provided")

        super().__init__(datadir=datadir, **kwargs)
        logger.debug("Initializing DataladDataGrabber")
        logger.debug(f"\turi = {uri}")
        logger.debug(f"\t_rootdir = {rootdir}")
        self.uri = uri
        self._rootdir = rootdir
        # Flag to indicate if the dataset was cloned before and it might be
        # dirty
        self.datalad_dirty = False

    @property
    def datadir(self) -> Path:
        """Get data directory path."""
        return super().datadir / self._rootdir

    def _get_dataset_id_remote(self) -> str:
        """Get the dataset id from the remote.

        Returns
        -------
        str
            The dataset id.

        """
        remote_id = None
        with tempfile.TemporaryDirectory() as tmpdir:
            logger.debug(f"Querying {self.uri} for dataset ID")
            repo = GitRepo.clone(
                self.uri, path=tmpdir, clone_options=["-n", "--depth=1"]
            )
            repo.checkout(name=".datalad/config", options=["HEAD"])
            remote_id = repo.config.get("datalad.dataset.id", None)
            logger.debug(f"Got remote dataset ID = {remote_id}")
        if remote_id is None:
            raise_error("Could not get dataset ID from remote")
        return remote_id

    def _dataset_get(self, out: Dict) -> Dict:
        """Get the dataset found from the path in `out`.

        Parameters
        ----------
        out : dict
            The dictionary from which path need to be searched.

        Returns
        -------
        dict
            The modified dictionary with meta updated.

        """
        to_get = [v["path"] for v in out.values() if "path" in v]

        if len(to_get) > 0:
            logger.debug(f"Getting {len(to_get)} files using datalad:")
            for fname in to_get:
                logger.debug(f"\t: {fname}")

            dl_out = self._dataset.get(to_get, result_renderer="disabled")
            if not self._was_cloned:
                # If the dataset was already installed, check that the
                # file was actually downloaded to avoid removing a
                # file that was already there.
                for t_out in dl_out:
                    t_path = Path(t_out["path"])
                    if t_out["status"] == "ok":
                        logger.debug(f"File {t_path} downloaded")
                        self._got_files.append(t_path)
                    elif t_out["status"] == "notneeded":
                        logger.debug(f"File {t_path} was already present")
                    else:
                        raise_error(f"File download failed: {t_out}")
            logger.debug("Get done")

        return out

    def install(self) -> None:
        """Install the datalad dataset into the datadir.

        Raises
        ------
        ValueError
            If the dataset is already installed but with a different ID.
        """
        isinstalled = dl.Dataset(self._datadir).is_installed()
        if isinstalled:
            logger.debug("Dataset already installed")
            self._got_files = []
            self._dataset: dl.Dataset = dl.Dataset(self._datadir)

            remote_id = self._get_dataset_id_remote()
            if remote_id != self._dataset.id:
                raise_error(
                    "Dataset already installed but with a different "
                    f"ID: {self._dataset.id} (local) != {remote_id} (remote)"
                )

            # Check for dirty datasets:
            status = self._dataset.status()
            if any([x["state"] != "clean" for x in status]):
                self.datalad_dirty = True
                warn_with_log(
                    "At least one file is not clean, Junifer will "
                    "consider this dataset as dirty."
                )
            else:
                logger.debug("Dataset is clean")

        else:
            logger.debug(f"Installing dataset {self.uri} to {self._datadir}")
            self._dataset: dl.Dataset = dl.clone(  # type: ignore
                self.uri, self._datadir, result_renderer="disabled"
            )
            logger.debug("Dataset installed")
        self._was_cloned = not isinstalled

        self.datalad_commit_id = self._dataset.repo.get_hexsha(  # type: ignore
            self._dataset.repo.get_corresponding_branch()  # type: ignore
        )
        self.datalad_id = self._dataset.id

    def cleanup(self) -> None:
        """Cleanup the datalad dataset."""
        if self._was_cloned:
            logger.debug("Removing dataset with reckless='kill'")
            self._dataset.remove(reckless="kill", result_renderer="disabled")
        else:
            logger.debug("Dropping files that were downloaded")
            for f in self._got_files:
                logger.debug(f"Dropping {f}")
                self._dataset.drop(f, result_renderer="disabled")

    def __getitem__(self, element: Union[str, Tuple]) -> Dict:
        """Implement single element indexing in the Datalad database.

        It will first obtain the paths from the parent class and then
        `datalad get` each of the files.

        This method only works with multiple inheritance.

        Parameters
        ----------
        element : str or tuple
            The element to be indexed. If one string is provided, it is
            assumed to be a tuple with only one item. If a tuple is provided,
            each item in the tuple is the value for the replacement string
            specified in "replacements".

        Returns
        -------
        dict
            Dictionary of paths for each type of data required for the
            specified element.

        """
        out = super().__getitem__(element)
        out = self._dataset_get(out)
        return out

    def __enter__(self):
        """Implement context entry."""
        self.install()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Implement context exit."""
        logger.debug("Cleaning up dataset")
        self.cleanup()
        logger.debug("Dataset state restored")



"""Provide class for AOMIC1000 VBM juseless datalad datagrabber."""

# Authors: Felix Hoffstaedter <f.hoffstaedter@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import Union

from ....api.decorators import register_datagrabber
from ....datagrabber import PatternDataladDataGrabber


@register_datagrabber
class JuselessDataladAOMICID1000VBM(PatternDataladDataGrabber):
    """Juseless AOMICID1000 VBM Data Grabber class.

    Implements a Data Grabber to access the AOMICID1000 VBM data in Juseless.

    Parameters
    ----------
    datadir : str or pathlib.Path, optional
        The directory where the datalad dataset will be cloned. If None,
        the datalad dataset will be cloned into a temporary directory
        (default None).

    """

    def __init__(self, datadir: Union[str, Path, None] = None) -> None:
        uri = "https://gin.g-node.org/felixh/ds003097_ReproVBM"
        types = ["VBM_GM"]
        replacements = ["subject"]
        patterns = {
            "VBM_GM": "sub-{subject}/mri/mwp1sub-{subject}_run-2_T1w.nii.gz",
        }
        super().__init__(
            types=types,
            datadir=datadir,
            uri=uri,
            replacements=replacements,
            patterns=patterns,
        )



"""

Extracting root sum of squares from edge-wise timeseries.
=========================================================
This example uses a ``RSSETSMarker`` to compute root sum of squares
of the edge-wise timeseries using the Schaefer parcellation
(100 rois and 200 rois, 17 Yeo networks) for a 4D nifti BOLD file.

Authors: Leonard Sasse, Sami Hamdan, Nicolas Nieto, Synchon Mandal

License: BSD 3 clause

"""

import tempfile

import junifer.testing.registry  # noqa: F401
from junifer.api import collect, run
from junifer.storage import SQLiteFeatureStorage
from junifer.utils import configure_logging


###############################################################################
# Set the logging level to info to see extra information:
configure_logging(level="INFO")

##############################################################################
# Define the datagrabber interface
datagrabber = {
    "kind": "SPMAuditoryTestingDatagrabber",
}

###############################################################################
# Define the markers interface
markers = [
    {
        "name": "Schaefer100x17_RSSETS",
        "kind": "RSSETSMarker",
        "parcellation": "Schaefer100x17",
    },
    {
        "name": "Schaefer200x17_RSSETS",
        "kind": "RSSETSMarker",
        "parcellation": "Schaefer200x17",
    },
]

###############################################################################
# Create a temporary directory for junifer feature extraction:
# At the end you can read the extracted data into a ``pandas.DataFrame``.
with tempfile.TemporaryDirectory() as tmpdir:
    # Define the storage interface
    storage = {
        "kind": "SQLiteFeatureStorage",
        "uri": f"{tmpdir}/test.sqlite",
    }
    # Run the defined junifer feature extraction pipeline
    run(
        workdir=tmpdir,
        datagrabber=datagrabber,
        markers=markers,
        storage=storage,
        elements=["sub001"],  # we calculate for one subject only
    )
    # Collect extracted features data
    collect(storage=storage)
    # Create storage object to read in extracted features
    db = SQLiteFeatureStorage(uri=storage["uri"])

    # List all the features
    print(db.list_features())
    # Read extracted features
    df_vbm = db.read_df(feature_name="BOLD_Schaefer100x17_RSSETS")

###############################################################################
# Now we take a look at the dataframe
df_vbm.head()



"""Provide tests for UCLA juseless datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import socket
from typing import Optional

import pytest

from junifer.configs.juseless.datagrabbers import JuselessUCLA
from junifer.utils.logging import configure_logging


# Check if the test is running on juseless
if socket.gethostname() != "juseless":
    pytest.skip("These tests are only for juseless", allow_module_level=True)

configure_logging(level="DEBUG")


def test_juseless_ucla_datagrabber() -> None:
    """Test juseless ucla datagrabber."""
    with JuselessUCLA() as dg:
        all_elements = dg.get_elements()
        test_element = all_elements[0]
        out = dg[test_element]

        types = [
            "BOLD",
            "BOLD_confounds",
            "T1w",
            "probseg_CSF",
            "probseg_GM",
            "probseg_WM",
        ]

        for t in types:
            assert t in out
            assert out[t]["path"].exists()


@pytest.mark.parametrize(
    "tasks",
    [None, "rest", ["rest", "stopsignal"]],
)
def test_juseless_ucla_datagrabber_task_params(tasks: Optional[str]) -> None:
    """Test juseless ucla datagrabber with different task parameters.

    Parameters
    ----------
    tasks : str
        The parametrized tasks in the UCLA dataset.

    """

    all_tasks = [
        "rest",
        "bart",
        "bht",
        "pamenc",
        "pamret",
        "scap",
        "taskswitch",
        "stopsignal",
    ]

    with JuselessUCLA(tasks=tasks) as dg:
        all_elements = dg.get_elements()

        if tasks is None:
            for el in all_elements:
                assert el[1] in all_tasks
        elif tasks == "rest":
            for el in all_elements:
                assert el[1] == "rest"
        else:
            for el in all_elements:
                assert el[1] in ["rest", "stopsignal"]


def test_juseless_ucla_datagrabber_invalid_tasks() -> None:
    """Test juseless ucla datagrabber with invalid task parameters."""
    with pytest.raises(
        ValueError, match="invalid is not a valid task in the UCLA"
    ):
        JuselessUCLA(tasks="invalid")



"""Provide utility functions for the api sub-package."""

# Authors: Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import os
import platform as pl
import re
from importlib.metadata import distribution
from typing import Dict

from .._version import __version__
from ..utils.logging import get_versions


def _get_junifer_version() -> Dict[str, str]:
    """Get junifer version information.

    Returns
    -------
    dict
        A dictionary containing junifer version.

    """
    return {
        "version": __version__,
    }


def _get_python_information() -> Dict[str, str]:
    """Get installed Python information.

    Parameters
    ----------
    dict
        A dictionary containing Python information.

    """
    return {
        "version": pl.python_version(),
        "implementation": pl.python_implementation(),
    }


def _get_dependency_information(long_: bool) -> Dict[str, str]:
    """Get Python environment dependency information.

    Parameters
    ----------
    long_ : bool
        Whether to report long version.

    Returns
    -------
    dict
        A dictionary containing Python dependency information.

    """
    dependency_versions = get_versions()

    pruned_dependency_versions = {}
    # Report long version
    if long_:
        for key, value in dependency_versions.items():
            # Ignore built-in modules and self
            if value != "None" and key != "junifer":
                pruned_dependency_versions[key] = value

    # Report short version
    else:
        # Get dependencies for junifer
        dist = distribution("junifer")
        # Compile regex pattern
        re_pattern = re.compile("[a-z-]+")

        for pkg_with_version in dist.requires:  # type: ignore
            # Perform regex search
            matches = re.findall(pattern=re_pattern, string=pkg_with_version)
            # Fix issue with PyYAML name registration
            if matches[0] == "pyyaml":
                key = "yaml"
            else:
                key = matches[0]

            if key in dependency_versions.keys():
                # Check if pkg part of optional dependencies
                if "extra" not in matches:
                    pruned_dependency_versions[key] = dependency_versions[key]

    return pruned_dependency_versions


def _get_system_information() -> Dict[str, str]:
    """Get system information.

    Returns
    -------
    dict
        A dictionary containing system information.

    """
    return {
        "platform": pl.platform(),
    }


def _get_environment_information(long_: bool) -> Dict[str, str]:
    """Get system environment information.

    Parameters
    ----------
    long_ : bool
        Whether to report long version.

    Returns
    -------
    dict
        A dictionary containing system environment information.

    """
    environment_values = {}
    # Report long version
    if long_:
        for key, value in os.environ.items():
            environment_values[key] = value
    # Report short version
    else:
        for key in ["LC_CTYPE", "LC_TERMINAL", "LC_TERMINAL_VERSION", "PATH"]:
            if key in os.environ.keys():
                environment_values[key] = os.environ[key]

    return environment_values



"""Provide tests for registry."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL
import logging
from abc import ABC
from typing import Type

import pytest

from junifer.datagrabber import PatternDataGrabber
from junifer.pipeline.registry import (
    build,
    get_class,
    get_step_names,
    register,
)
from junifer.storage import SQLiteFeatureStorage


def test_register_invalid_step():
    """Test register invalid step name."""
    with pytest.raises(ValueError, match="Invalid step:"):
        register(step="foo", name="bar", klass=str)


# TODO: improve parametrization
@pytest.mark.parametrize(
    "step, name, klass",
    [
        ("datagrabber", "pattern-dg", PatternDataGrabber),
        ("storage", "sqlite-storage", SQLiteFeatureStorage),
    ],
)
def test_register(
    caplog: pytest.LogCaptureFixture, step: str, name: str, klass: Type
) -> None:
    """Test register.

    Parameters
    ----------
    caplog : pytest.LogCaptureFixture
        A pytest fixture to capture logging.
    step : str
        The parametrized name of the step.
    name : str
        The parametrized name of the function.
    klass : str
        The parametrized name of the base class.

    """
    with caplog.at_level(logging.INFO):
        # Register
        register(step=step, name=name, klass=klass)
        # Check logging message
        assert "Registering" in caplog.text


def test_get_step_names_invalid_step() -> None:
    """Test get step name invalid step name."""
    with pytest.raises(ValueError, match="Invalid step:"):
        get_step_names(step="foo")


def test_get_step_names_absent() -> None:
    """Test get step names for absent name."""
    # Get step names for datagrabber
    datagrabbers = get_step_names(step="datagrabber")
    # Check for datagrabber step name
    assert "bar" not in datagrabbers


def test_get_step_names() -> None:
    """Test get step names."""
    # Register datagrabber
    register(step="datagrabber", name="bar", klass=str)
    # Get step names for datagrabber
    datagrabbers = get_step_names(step="datagrabber")
    # Check for datagrabber step name
    assert "bar" in datagrabbers


def test_get_class_invalid_step() -> None:
    """Test get class invalid step name."""
    with pytest.raises(ValueError, match="Invalid step:"):
        get_class(step="foo", name="bar")


def test_get_class_invalid_name() -> None:
    """Test get class invalid function name."""
    with pytest.raises(ValueError, match="Invalid name:"):
        get_class(step="datagrabber", name="foo")


# TODO: enable parametrization
def test_get_class():
    """Test get class."""
    # Register datagrabber
    register(step="datagrabber", name="bar", klass=str)
    # Get class
    obj = get_class(step="datagrabber", name="bar")
    assert obj == str


# TODO: possible parametrization?
def test_build():
    """Test building objects from names."""
    import numpy as np

    # Define abstract base class
    class SuperClass(ABC):
        pass

    # Define concrete class
    class ConcreteClass(SuperClass):
        def __init__(self, value=1):
            self.value = value

    # Register
    register(step="datagrabber", name="concrete", klass=ConcreteClass)

    # Build
    obj = build(step="datagrabber", name="concrete", baseclass=SuperClass)
    assert isinstance(obj, ConcreteClass)
    assert obj.value == 1

    # Build
    obj = build(
        step="datagrabber",
        name="concrete",
        baseclass=SuperClass,
        init_params={"value": 2},
    )
    assert isinstance(obj, ConcreteClass)
    assert obj.value == 2

    # Check error
    with pytest.raises(ValueError, match="Must inherit"):
        build(step="datagrabber", name="concrete", baseclass=np.ndarray)



"""Provide tests for cli."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import Tuple

import pytest
import yaml
from click.testing import CliRunner

from junifer.api.cli import collect, run, selftest, wtf


# Create click test runner
runner = CliRunner()


# TODO: adapt elements to take arrays
@pytest.mark.parametrize(
    "elements",
    [
        ("sub-01", "sub-02", "sub-03"),
        ("sub-01", "sub-02", "sub-04"),
    ],
)
def test_run_and_collect_commands(
    tmp_path: Path, elements: Tuple[str, ...]
) -> None:
    """Test run and collect commands."""
    # Get test config
    infile = Path(__file__).parent / "data" / "gmd_mean.yaml"
    # Read test config
    with open(infile, mode="r") as f:
        contents = yaml.safe_load(f)
    # Working directory
    workdir = tmp_path / "workdir"
    contents["workdir"] = str(workdir.absolute())
    # Output directory
    outdir = tmp_path / "outdir"
    # Storage
    contents["storage"]["uri"] = str(outdir.absolute())
    # Write new test config
    outfile = tmp_path / "in.yaml"
    with open(outfile, mode="w") as f:
        yaml.dump(contents, f)
    # Run command arguments
    run_args = [
        str(outfile.absolute()),
        "--verbose",
        "debug",
        "--element",
        elements[0],
        "--element",
        elements[1],
        "--element",
        elements[2],
    ]
    # Invoke run command
    run_result = runner.invoke(run, run_args)
    # Check
    assert run_result.exit_code == 0
    # Collect command arguments
    collect_args = [str(outfile.absolute()), "--verbose", "debug"]
    # Invoke collect command
    collect_result = runner.invoke(collect, collect_args)
    # Check
    assert collect_result.exit_code == 0


def test_wtf_short() -> None:
    """Test short version of wtf command."""
    # Invoke wtf command
    wtf_result = runner.invoke(wtf)
    # Check
    assert wtf_result.exit_code == 0


def test_wtf_long() -> None:
    """Test long version of wtf command."""
    # Invoke wtf command
    wtf_result = runner.invoke(wtf, "--long")
    # Check
    assert wtf_result.exit_code == 0


def test_selftest_invalid_arg() -> None:
    """Test selftest failure for invalid argument value."""
    # Invoke selftest command
    selftest_result = runner.invoke(selftest, "abyss")
    # Check
    assert selftest_result.exit_code == 2


def test_selftest() -> None:
    """Test selftest."""
    # Invoke selftest command
    selftest_result = runner.invoke(selftest, "tests")
    # Check; will result in 1 due to I/O descriptor manipulation
    assert selftest_result.exit_code == 1



"""Provide GitHub substitutions in Sphinx."""

from docutils.nodes import reference
from docutils.parsers.rst.roles import set_classes


# Taken from MNE-Python
# adapted from
# https://doughellmann.com/blog/2010/05/09/defining-custom-roles-in-sphinx/


def gh_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """Link to a GitHub issue."""
    try:
        # issue/PR mode (issues/PR-num will redirect to pull/PR-num)
        int(text)
    except ValueError:
        # direct link mode
        slug = text
    else:
        slug = "issues/" + text
    text = "#" + text
    ref = "https://github.com/juaml/junifer/" + slug
    set_classes(options)
    node = reference(rawtext, text, refuri=ref, **options)
    return [node], []


def setup(app):
    app.add_role("gh", gh_role)
    return



"""Provide class for marker collection."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from collections import Counter
from typing import TYPE_CHECKING, Dict, List, Optional

from ..datareader.default import DefaultDataReader
from ..markers.base import BaseMarker
from ..pipeline import PipelineStepMixin
from ..storage.base import BaseFeatureStorage
from ..utils import logger


if TYPE_CHECKING:
    from junifer.datagrabber import BaseDataGrabber


class MarkerCollection:
    """Class for marker collection.

    Parameters
    ----------
    markers : list of marker-like
        The markers to compute.
    datareader : datareader-like, optional
        The datareader to use (default None).
    preprocessing : preprocessing-like, optional
        The preprocessing steps to apply.
    storage : storage-like, optional
        The storage to use (default None).

    """

    def __init__(
        self,
        markers: List[BaseMarker],
        datareader: Optional[PipelineStepMixin] = None,
        preprocessing: Optional[PipelineStepMixin] = None,
        storage: Optional[BaseFeatureStorage] = None,
    ):
        # Check that the markers have different names
        marker_names = [m.name for m in markers]
        if len(set(marker_names)) != len(marker_names):
            counts = Counter(marker_names)
            raise ValueError(
                "Markers must have different names. "
                f"Current names are: {counts}"
            )
        self._markers = markers
        if datareader is None:
            datareader = DefaultDataReader()
        self._datareader = datareader
        self._preprocessing = preprocessing
        self._storage = storage

    def fit(self, input: Dict[str, Dict]) -> Optional[Dict]:
        """Fit the pipeline.

        Parameters
        ----------
        input : dict
            The input data to fit the pipeline on. Should be the output of
            indexing the Data Grabber with one element.

        Returns
        -------
        dict or None
            The output of the pipeline. Each key represents a marker name and
            the values are the computer marker values. If the pipeline has a
            storage configured, then the output will be None.

        """
        logger.info("Fitting pipeline")

        # Fetch actual data using datareader
        data = self._datareader.fit_transform(input)

        # Apply preprocessing steps
        if self._preprocessing is not None:
            logger.info("Preprocessing data")
            data = self._preprocessing.fit_transform(data)

        # Compute markers
        out = {}
        for marker in self._markers:
            logger.info(f"Fitting marker {marker.name}")
            m_value = marker.fit_transform(data, storage=self._storage)
            if self._storage is None:
                out[marker.name] = m_value
        logger.info("Marker collection fitting done")

        return None if self._storage else out

    def validate(self, datagrabber: "BaseDataGrabber") -> None:
        """Validate the pipeline.

        Without doing any computation, check if the marker collection can
        be fit without problems. That is, the data required for each marker is
        present and streamed down the steps. Also, if a storage is configured,
        check that the storage can handle the markers output.

        Parameters
        ----------
        datagrabber : datagrabber-like
            The datagrabber to validate.

        """
        logger.info("Validating Marker Collection")
        t_data = datagrabber.get_types()
        logger.info(f"DataGrabber output type: {t_data}")

        logger.info("Validating Data Reader:")
        t_data = self._datareader.validate(t_data)
        logger.info(f"Data Reader output type: {t_data}")

        for marker in self._markers:
            logger.info(f"Validating Marker: {marker.name}")
            # Validate marker
            m_data = marker.validate(input=t_data)
            logger.info(f"Marker output type: {m_data}")
            # Check storage for the marker
            if self._storage is not None:
                logger.info(f"Validating storage for {marker.name}")
                # Validate storage
                self._storage.validate(input_=m_data)



"""Provide tests for PartlyCloudy datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from junifer.testing.datagrabbers import PartlyCloudyTestingDataGrabber


def test_PartlyCloudyTestingDataGrabber() -> None:
    """Test PartlyCloudy datagrabber."""
    expected_elements = [
        "sub-01",
        "sub-02",
        "sub-03",
        "sub-04",
        "sub-05",
        "sub-06",
        "sub-07",
        "sub-08",
        "sub-09",
        "sub-10",
    ]
    with PartlyCloudyTestingDataGrabber() as dg:
        all_elements = dg.get_elements()
        assert set(all_elements) == set(expected_elements)
        out = dg["sub-01"]
        assert "BOLD" in out
        assert out["BOLD"]["path"].exists()
        assert out["BOLD"]["path"].is_file()

        assert "BOLD_confounds" in out
        assert out["BOLD_confounds"]["path"].exists()
        assert out["BOLD_confounds"]["path"].is_file()
        assert "format" in out["BOLD_confounds"]
        assert "fmriprep" == out["BOLD_confounds"]["format"]

    with PartlyCloudyTestingDataGrabber(reduce_confounds=False) as dg:
        out = dg["sub-01"]
        assert "format" in out["BOLD_confounds"]
        assert "fmriprep" == out["BOLD_confounds"]["format"]



"""Provide imports for datareader sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from .default import DefaultDataReader



"""Provide tests for testing registry."""

import importlib

from junifer.pipeline.registry import get_step_names


def test_testing_registry() -> None:
    """Test testing registry."""
    import junifer

    importlib.reload(junifer.pipeline.registry)
    importlib.reload(junifer)

    assert "OasisVBMTestingDatagrabber" not in get_step_names("datagrabber")
    assert "SPMAuditoryTestingDatagrabber" not in get_step_names("datagrabber")
    importlib.reload(junifer.testing.registry)  # type: ignore

    assert "OasisVBMTestingDatagrabber" in get_step_names("datagrabber")
    assert "SPMAuditoryTestingDatagrabber" in get_step_names("datagrabber")



"""Provide functions for statistics."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Any, Callable, Dict, Optional

import numpy as np
from scipy.stats import trim_mean
from scipy.stats.mstats import winsorize

from .utils import logger, raise_error


def get_aggfunc_by_name(
    name: str, func_params: Optional[Dict[str, Any]] = None
) -> Callable:
    """Get an aggregation function by its name.

    Parameters
    ----------
    name : str
        Name to identify the function. Currently supported names and
        corresponding functions are:

        * ``winsorized_mean`` -> :func:`scipy.stats.mstats.winsorize`
        * ``mean`` -> :func:`numpy.mean`
        * ``std`` -> :func:`numpy.std`
        * ``trim_mean`` -> :func:`scipy.stats.trim_mean`

    func_params : dict, optional
        Parameters to pass to the function.
        E.g. for ``winsorized_mean``: ``func_params = {'limits': [0.1, 0.1]}``
        (default None).

    Returns
    -------
    function
        Respective function with ``func_params`` parameter set.
    """
    from functools import partial  # local import to avoid sphinx error

    # check validity of names
    _valid_func_names = {"winsorized_mean", "mean", "std", "trim_mean"}
    if func_params is None:
        func_params = {}
    # apply functions
    if name == "winsorized_mean":
        # check validity of func_params
        limits = func_params.get("limits")
        if limits is None or not isinstance(limits, list):
            raise_error(
                "func_params must contain a list of limits for "
                "winsorized_mean",
                ValueError,
            )
        if len(limits) != 2:
            raise_error(
                "func_params must contain a list of two limits for "
                "winsorized_mean",
                ValueError,
            )
        if any((lim < 0.0 or lim > 1) for lim in limits):
            raise_error(
                "Limits for the winsorized mean must be between 0 and 1."
            )
        logger.info(f"Limits for winsorized mean are set to {limits}.")
        # partially interpret func_params
        func = partial(winsorized_mean, **func_params)
    elif name == "mean":
        func = np.mean
    elif name == "std":
        func = np.std
    elif name == "trim_mean":
        func = partial(trim_mean, **func_params)
    else:
        raise_error(
            f"Function {name} unknown. Please provide any of "
            f"{_valid_func_names}"
        )
    return func


def winsorized_mean(
    data: np.ndarray, axis: Optional[int] = None, **win_params
) -> np.ndarray:
    """Compute a winsorized mean by chaining winsorization and mean.

    Parameters
    ----------
    data : numpy.ndarray
        Data to calculate winsorized mean on.
    axis : int, optional
        The axis to calculate winsorized mean on (default None).
    **win_params : dict
        Dictionary containing the keyword arguments for the winsorize function.
        E.g. ``{'limits': [0.1, 0.1]}``.

    Returns
    -------
    numpy.ndarray
        Winsorized mean of the inputted data with the winsorize settings
        applied as specified in ``win_params``.

    See Also
    --------
    scipy.stats.mstats.winsorize :
        The winsorize function used in this function.
    """
    win_dat = winsorize(data, axis=axis, **win_params)
    win_mean = win_dat.mean(axis=axis)

    return win_mean



"""Provide tests for base."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path

import pytest

from junifer.datagrabber.base import BaseDataGrabber


def test_BaseDataGrabber_abstractness() -> None:
    """Test BaseDataGrabber is abstract base class."""
    with pytest.raises(TypeError, match=r"abstract"):
        BaseDataGrabber(datadir="/tmp", types=["func"])  # type: ignore


def test_BaseDataGrabber() -> None:
    """Test BaseDataGrabber."""
    # Create concrete class.
    class MyDataGrabber(BaseDataGrabber):
        def get_item(self, subject):
            return {"BOLD": {}}

        def get_elements(self):
            return super().get_elements()

        def get_element_keys(self):
            return ["subject"]

    dg = MyDataGrabber(datadir="/tmp", types=["BOLD"])
    elem = dg["sub01"]
    assert "BOLD" in elem
    assert "meta" in elem["BOLD"]
    meta = elem["BOLD"]["meta"]
    assert "datagrabber" in meta
    assert "class" in meta["datagrabber"]
    assert MyDataGrabber.__name__ in meta["datagrabber"]["class"]
    assert "element" in meta
    assert "subject" in meta["element"]
    assert "sub01" in meta["element"]["subject"]

    with pytest.raises(NotImplementedError):
        dg.get_elements()

    with dg:
        assert dg.datadir == Path("/tmp")
        assert dg.types == ["BOLD"]

    class MyDataGrabber2(BaseDataGrabber):
        def get_item(self, subject):
            return super().get_item(subject=subject)

        def get_elements(self):
            return super().get_elements()

        def get_element_keys(self):
            return super().get_element_keys()

    dg = MyDataGrabber2(datadir="/tmp", types=["BOLD"])
    with pytest.raises(NotImplementedError):
        dg.get_element_keys()

    with pytest.raises(NotImplementedError):
        dg.get_item(subject=1)  # type: ignore



"""Provide tests for Oasis VBM Testing datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from junifer.testing.datagrabbers import OasisVBMTestingDatagrabber


def test_OasisVBMTestingDatagrabber() -> None:
    """Test Oasis VBM Testing datagrabber."""
    expected_elements = [
        "sub-01",
        "sub-02",
        "sub-03",
        "sub-04",
        "sub-05",
        "sub-06",
        "sub-07",
        "sub-08",
        "sub-09",
        "sub-10",
    ]
    with OasisVBMTestingDatagrabber() as dg:
        all_elements = dg.get_elements()
        assert set(all_elements) == set(expected_elements)
        out = dg["sub-01"]
        assert "VBM_GM" in out
        assert out["VBM_GM"]["path"].exists()
        assert out["VBM_GM"]["path"].is_file()



"""Provide tests for logging."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Sami Hamdan <s.hamdan@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import logging
from pathlib import Path

import pytest

from junifer.utils.logging import (
    _close_handlers,
    configure_logging,
    get_versions,
    log_versions,
    logger,
    raise_error,
    warn_with_log,
)


def test_get_versions() -> None:
    """Test version info fetch for modules."""
    module_versions = get_versions()
    assert "junifer" in module_versions.keys()


def test_log_versions(caplog: pytest.LogCaptureFixture) -> None:
    """Test logging of dependency and junifer versions.

    Parameters
    ----------
    caplog : pytest.LogCaptureFixture
        A pytest fixture to capture logging.

    """
    # Set log capturing at INFO
    with caplog.at_level(logging.INFO):
        # Log versions
        log_versions()
        # Check logging levels
        for record in caplog.records:
            assert record.levelname not in ("DEBUG", "WARNING", "ERROR")
            assert record.levelname == "INFO"
        # Check logging message
        assert "junifer" in caplog.text


def test_log_file(tmp_path: Path) -> None:
    """Test logging to a file.

    tmp_path : Path
        The path to the test directory.

    """
    configure_logging(fname=tmp_path / "test1.log")
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warn message")
    logger.error("Error message")
    _close_handlers(logger)
    with open(tmp_path / "test1.log") as f:
        lines = f.readlines()
        assert not any("Debug message" in line for line in lines)
        assert not any("Info message" in line for line in lines)
        assert any("Warn message" in line for line in lines)
        assert any("Error message" in line for line in lines)

    configure_logging(fname=tmp_path / "test2.log", level="INFO")
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warn message")
    logger.error("Error message")
    _close_handlers(logger)
    with open(tmp_path / "test2.log") as f:
        lines = f.readlines()
        assert not any("Debug message" in line for line in lines)
        assert any("Info message" in line for line in lines)
        assert any("Warn message" in line for line in lines)
        assert any("Error message" in line for line in lines)

    configure_logging(fname=tmp_path / "test3.log", level="WARNING")
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warn message")
    logger.error("Error message")
    _close_handlers(logger)
    with open(tmp_path / "test3.log") as f:
        lines = f.readlines()
        assert not any("Debug message" in line for line in lines)
        assert not any("Info message" in line for line in lines)
        assert any("Warn message" in line for line in lines)
        assert any("Error message" in line for line in lines)

    configure_logging(fname=tmp_path / "test4.log", level="ERROR")
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warn message")
    logger.error("Error message")
    with open(tmp_path / "test4.log") as f:
        lines = f.readlines()
        assert not any("Debug message" in line for line in lines)
        assert not any("Info message" in line for line in lines)
        assert not any("Warn message" in line for line in lines)
        assert any("Error message" in line for line in lines)

    with pytest.warns(UserWarning, match="to avoid this message"):
        configure_logging(fname=tmp_path / "test4.log", level="WARNING")
        logger.debug("Debug2 message")
        logger.info("Info2 message")
        logger.warning("Warn2 message")
        logger.error("Error2 message")
        with open(tmp_path / "test4.log") as f:
            lines = f.readlines()
            assert not any("Debug message" in line for line in lines)
            assert not any("Info message" in line for line in lines)
            assert not any("Warn message" in line for line in lines)
            assert any("Error message" in line for line in lines)
            assert not any("Debug2 message" in line for line in lines)
            assert not any("Info2 message" in line for line in lines)
            assert any("Warn2 message" in line for line in lines)
            assert any("Error2 message" in line for line in lines)

    configure_logging(
        fname=tmp_path / "test4.log", level="WARNING", overwrite=True
    )
    logger.debug("Debug3 message")
    logger.info("Info3 message")
    logger.warning("Warn3 message")
    logger.error("Error3 message")
    with open(tmp_path / "test4.log") as f:
        lines = f.readlines()
        assert not any("Debug message" in line for line in lines)
        assert not any("Info message" in line for line in lines)
        assert not any("Warn message" in line for line in lines)
        assert not any("Error message" in line for line in lines)
        assert not any("Debug2 message" in line for line in lines)
        assert not any("Info2 message" in line for line in lines)
        assert not any("Warn2 message" in line for line in lines)
        assert not any("Error2 message" in line for line in lines)
        assert not any("Debug3 message" in line for line in lines)
        assert not any("Info3 message" in line for line in lines)
        assert any("Warn3 message" in line for line in lines)
        assert any("Error3 message" in line for line in lines)

    with pytest.warns(RuntimeWarning, match=r"Warn raised"):
        warn_with_log("Warn raised")
    with pytest.raises(ValueError, match=r"Error raised"):
        raise_error("Error raised")
    with open(tmp_path / "test4.log") as f:
        lines = f.readlines()
        assert any("Warn raised" in line for line in lines)
        assert any("Error raised" in line for line in lines)


def test_log_stdout(caplog: pytest.LogCaptureFixture) -> None:
    """Test logging to stdout.

    Parameters
    ----------
    caplog : pytest.LogCaptureFixture
        A pytest fixture to capture logging.

    """
    configure_logging()
    logger.info("Testing")
    for record in caplog.records:
        assert record.levelname == "INFO"


def test_lib_logging(tmp_path: Path) -> None:
    """Test logging versions.

    tmp_path : Path
        The path to the test directory.

    """
    # Import third-party packages
    import numpy as np  # noqa
    import pandas  # noqa

    log_file_path = tmp_path / "test_lib_logging.log"
    configure_logging(fname=log_file_path, level="INFO")
    logger.info("first message")
    with open(log_file_path) as f:
        lines = f.readlines()
        assert any("numpy" in line for line in lines)
        assert any("pandas" in line for line in lines)
        assert any("junifer" in line for line in lines)


def test_raise_error(caplog: pytest.LogCaptureFixture) -> None:
    """Test logging and raising error.

    Parameters
    ----------
    caplog : pytest.LogCaptureFixture
        A pytest fixture to capture logging.

    """
    with pytest.raises(ValueError, match="test error"):
        raise_error(msg="test error")
        for record in caplog.records:
            assert record.levelname == "ERROR"


def test_warn_with_log(caplog: pytest.LogCaptureFixture) -> None:
    """Test logging and warning.

    Parameters
    ----------
    caplog : pytest.LogCaptureFixture
        A pytest fixture to capture logging.

    """
    with pytest.warns(RuntimeWarning, match="test warning"):
        warn_with_log("test warning")
        for record in caplog.records:
            assert record.levelname == "WARNING"



"""Provide tests for update meta mixin."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Dict, List, Set, Union

import pytest

from junifer.pipeline.update_meta_mixin import UpdateMetaMixin


@pytest.mark.parametrize(
    "input, step_name, dependencies, expected",
    [
        ({}, "step_name", None, set()),
        ({}, "step_name", ["numpy"], set(["numpy"])),
        ({}, "step_name", "numpy", set(["numpy"])),
        ({}, "step_name", set(["numpy"]), set(["numpy"])),
    ],
)
def test_UpdateMetaMixin(
    input: Dict,
    step_name: str,
    dependencies: Union[Set, List, str, None],
    expected: Set,
) -> None:
    """Test UpdateMetaMixin.

    Parameters
    ----------
    input : dict
        The data object to update.
    step_name : str
        The name of the pipeline step.
    dependencies : set, list, None, str
        The dependencies of the pipeline step.
    expected : set
        The expected dependencies.
    """

    class TestUpdateMetaMixin(UpdateMetaMixin):
        """Test UpdateMetaMixin."""

        _DEPENDENCIES = dependencies

    obj = TestUpdateMetaMixin()
    obj.update_meta(input, step_name=step_name)
    assert input["meta"]["step_name"]["class"] == "TestUpdateMetaMixin"
    assert input["meta"]["dependencies"] == expected



"""Testing utils."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from pathlib import Path


def get_testing_data(fname: str) -> Path:
    """Get the path to a testing data file.

    Parameters
    ----------
    fname : str
        The name of the file.

    Returns
    -------
    pathlib.Path
        The absolute path to the file.
    """
    t_path = Path(__file__).parent / "data" / fname
    if not t_path.exists():
        raise FileNotFoundError(f"File {fname} not found in testing data")
    return t_path.resolve()



"""Provide imports for testing sub-package."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from . import datagrabbers
from .utils import get_testing_data



"""
UKB VBM GMD Extraction
======================

Authors: Federico Raimondo

License: BSD 3 clause
"""


from junifer.api import run


markers = [
    {
        "name": "Schaefer1000x7_TrimMean80",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer1000x7",
        "method": "trim_mean",
        "method_params": {"proportiontocut": 0.2},
    },
    {
        "name": "Schaefer1000x7_Mean",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer1000x7",
        "method": "mean",
    },
    {
        "name": "Schaefer1000x7_Std",
        "kind": "ParcelAggregation",
        "parcellation": "Schaefer1000x7",
        "method": "std",
    },
]

run(
    workdir="/tmp",
    datagrabber="JuselessUKBVBM",
    elements=("sub-1627474", "ses-2"),
    markers=markers,
    storage="SQLDataFrameStorage",
    storage_params={"outpath": "/data/project/juniferexample"},
)



"""Provide class for parcel aggregation."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import Any, Dict, List, Optional, Union

import numpy as np
from nilearn.image import math_img, new_img_like, resample_to_img
from nilearn.maskers import NiftiMasker

from ..api.decorators import register_marker
from ..data import load_mask, load_parcellation
from ..stats import get_aggfunc_by_name
from ..utils import logger
from .base import BaseMarker


@register_marker
class ParcelAggregation(BaseMarker):
    """Class for parcel aggregation.

    Parameters
    ----------
    parcellation : str or list of str
        The name(s) of the parcellation(s). Check valid options by calling
        :func:`junifer.data.parcellations.list_parcellations`.
    method : str
        The method to perform aggregation using. Check valid options in
        :func:`junifer.stats.get_aggfunc_by_name`.
    method_params : dict, optional
        Parameters to pass to the aggregation function. Check valid options in
        :func:`junifer.stats.get_aggfunc_by_name`.
    mask : str, optional
        The name of the mask to apply to regions before extracting signals.
        Check valid options by calling :func:`junifer.data.masks.list_masks`
        (default None).
    on : {"T1w", "BOLD", "VBM_GM", "VBM_WM", "fALFF", "GCOR", "LCOR"} \
         or list of the options, optional
        The data types to apply the marker to. If None, will work on all
        available data (default None).
    name : str, optional
        The name of the marker. If None, will use the class name (default
        None).
    """

    _DEPENDENCIES = {"nilearn", "numpy"}

    def __init__(
        self,
        parcellation: Union[str, List[str]],
        method: str,
        method_params: Optional[Dict[str, Any]] = None,
        mask: Optional[str] = None,
        on: Union[List[str], str, None] = None,
        name: Optional[str] = None,
    ) -> None:
        if not isinstance(parcellation, list):
            parcellation = [parcellation]
        self.parcellation = parcellation
        self.method = method
        self.method_params = method_params or {}
        self.mask = mask
        super().__init__(on=on, name=name)

    def get_valid_inputs(self) -> List[str]:
        """Get valid data types for input.

        Returns
        -------
        list of str
            The list of data types that can be used as input for this marker.

        """
        return ["T1w", "BOLD", "VBM_GM", "VBM_WM", "fALFF", "GCOR", "LCOR"]

    def get_output_type(self, input_type: str) -> str:
        """Get output type.

        Parameters
        ----------
        input_type : str
            The data type input to the marker.

        Returns
        -------
        str
            The storage type output by the marker.

        """

        if input_type in ["VBM_GM", "VBM_WM", "fALFF", "GCOR", "LCOR"]:
            return "table"
        elif input_type == "BOLD":
            return "timeseries"
        else:
            raise ValueError(f"Unknown input kind for {input_type}")

    def compute(
        self,
        input: Dict[str, Any],
        extra_input: Optional[Dict] = None,
    ) -> Dict:
        """Compute.

        Parameters
        ----------
        input : dict
            A single input from the pipeline data object in which to compute
            the marker.
        extra_input : dict, optional
            The other fields in the pipeline data object. Useful for accessing
            other data kind that needs to be used in the computation. For
            example, the functional connectivity markers can make use of the
            confounds if available (default None).

        Returns
        -------
        dict
            The computed result as dictionary. This will be either returned
            to the user or stored in the storage by calling the store method
            with this as a parameter. The dictionary has the following keys:

            * ``data`` : the actual computed values as a numpy.ndarray
            * ``columns`` : the column labels for the computed values as a list
            * ``row_names`` (if more than one row is present in data): "scan"

        """
        t_input = input["data"]
        logger.debug(f"Parcel aggregation using {self.method}")
        agg_func = get_aggfunc_by_name(
            name=self.method,
            func_params=self.method_params,
        )
        # Get the min of the voxels sizes and use it as the resolution
        resolution = np.min(t_input.header.get_zooms()[:3])

        # Load the parcellations
        all_parcelations = []
        all_labels = []
        for t_parc_name in self.parcellation:
            t_parcellation, t_labels, _ = load_parcellation(
                name=t_parc_name,
                resolution=resolution,
            )
            # Resample all of them to the image
            t_parcellation_img_res = resample_to_img(
                t_parcellation,
                t_input,
                interpolation="nearest",
                copy=True,
            )
            all_parcelations.append(t_parcellation_img_res)
            all_labels.append(t_labels)

        # Avoid merging if there is only one parcellation
        if len(all_parcelations) == 1:
            parcellation_img_res = all_parcelations[0]
            labels = all_labels[0]
        else:
            # Merge the parcellations
            parc_data = all_parcelations[0].get_fdata()
            labels = all_labels[0]
            for t_parc, t_labels in zip(all_parcelations[1:], all_labels[1:]):
                # Get the data from this parcellation
                t_parc_data = t_parc.get_fdata().copy()  # must be copied
                # Increase the values of each ROI to match the labels
                t_parc_data[t_parc_data != 0] += len(labels)

                # Only set new values for the voxels that are 0
                # This makes sure that the voxels that are in multiple
                # parcellations are assigned to the parcellation that was
                # first in the list.
                parc_data[parc_data == 0] += t_parc_data[parc_data == 0]
                labels.extend(t_labels)

            parcellation_img_res = new_img_like(
                all_parcelations[0],
                parc_data,
            )

        parcellation_bin = math_img(
            "img != 0",
            img=parcellation_img_res,
        )

        if self.mask is not None:
            logger.debug(f"Masking with {self.mask}")
            mask_img, _ = load_mask(name=self.mask, resolution=resolution)
            mask_img = resample_to_img(
                mask_img,
                t_input,
                interpolation="nearest",
                copy=True,
            )
            parcellation_bin = math_img(
                "np.logical_and(img, mask)",
                img=parcellation_bin,
                mask=mask_img,
            )

        logger.debug("Masking")
        masker = NiftiMasker(
            parcellation_bin, target_affine=t_input.affine
        )  # type: ignore

        # Mask the input data and the parcellation
        data = masker.fit_transform(t_input)
        parcellation_values = masker.transform(parcellation_img_res)
        parcellation_values = np.squeeze(parcellation_values).astype(int)

        # Get the values for each parcel and apply agg function
        logger.debug("Computing ROI means")
        parcellation_roi_vals = sorted(np.unique(parcellation_values))
        out_labels = []
        out_values = []
        # Iterate over the parcels (existing)
        for t_v in parcellation_roi_vals:
            t_values = agg_func(data[:, parcellation_values == t_v], axis=-1)
            out_values.append(t_values)
            # Update the labels just in case a parcel has no voxels
            # in it
            out_labels.append(labels[t_v - 1])

        out_values = np.array(out_values).T
        out = {"data": out_values, "columns": out_labels}
        return out



"""Provide tests for default data reader."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path

import nibabel as nib
import pandas as pd
import pytest
from nibabel import testing as nib_testing
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

from junifer.datareader import DefaultDataReader


@pytest.mark.parametrize(
    "type_", [["T1w", "BOLD", "T2", "dwi"], [], ["whatever"]]
)
def test_validation(type_) -> None:
    """Test validating input/output.

    Parameters
    ----------
    type_ : list of str or str or None
        The parametrized type_ of data.

    """
    reader = DefaultDataReader()
    assert reader.validate_input(type_) is None
    assert reader.get_output_type(type_) == type_
    assert reader.validate(type_) == type_


def test_meta() -> None:
    """Test reader metadata."""
    reader = DefaultDataReader()

    nib_data_path = Path(nib_testing.data_path)
    t_path = nib_data_path / "example4d.nii.gz"
    input = {"BOLD": {"path": t_path}}
    output = reader.fit_transform(input)
    assert "meta" in output["BOLD"]
    meta = output["BOLD"]["meta"]
    assert "datareader" in meta
    assert "class" in meta["datareader"]
    assert meta["datareader"]["class"] == "DefaultDataReader"


@pytest.mark.parametrize(
    "fname", ["example4d.nii.gz", "reoriented_anat_moved.nii"]
)
def test_read_nifti(fname: str) -> None:
    """Test reading NIFTI files.

    Parameters
    ----------
    fname : str
        The parametrized NIfTI file names for testing.

    """
    reader = DefaultDataReader()
    nib_data_path = Path(nib_testing.data_path)

    t_path = nib_data_path / fname

    input = {"BOLD": {"path": t_path}}
    output = reader.fit_transform(input)

    assert isinstance(output, dict)
    assert "BOLD" in output
    assert isinstance(output["BOLD"], dict)
    assert "path" in output["BOLD"]
    assert "data" in output["BOLD"]

    read_img = output["BOLD"]["data"]

    t_read_img = nib.load(t_path)
    assert_array_equal(read_img.get_fdata(), t_read_img.get_fdata())

    input = {"BOLD": {"path": t_path.as_posix()}}
    output2 = reader.fit_transform(input)
    assert output["BOLD"]["path"] == output2["BOLD"]["path"]


def test_read_unknown() -> None:
    """Test (not) reading unknown files."""
    reader = DefaultDataReader()
    nib_data_path = Path(nib_testing.data_path)

    anat_path = nib_data_path / "reoriented_anat_moved.nii"
    whatever_path = nib_data_path / "unexistent.unkwnownextension"

    input = {"anat": {"path": anat_path}, "whatever": {"path": whatever_path}}
    output = reader.fit_transform(input)

    assert isinstance(output, dict)
    assert "anat" in output
    assert isinstance(output["anat"], dict)
    assert "path" in output["anat"]
    assert isinstance(output["anat"]["path"], Path)
    assert "data" in output["anat"]
    assert output["anat"]["data"] is not None

    assert isinstance(output["whatever"], dict)
    assert "path" in output["whatever"]
    assert isinstance(output["whatever"]["path"], Path)
    assert "data" in output["whatever"]
    assert output["whatever"]["data"] is None

    input = {"whatever": {"nopath": whatever_path}}
    with pytest.warns(RuntimeWarning, match="does not provide a path"):
        reader.fit_transform(input)


def test_read_csv(tmp_path: Path) -> None:
    """Test reading CSV files.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the test directory.

    """
    d = {"col1": [1, 2, 3, 4, 5], "col2": [3, 4, 5, 6, 7]}
    df = pd.DataFrame(d)

    df.to_csv(tmp_path / "test_read_csv.csv")

    reader = DefaultDataReader()
    input = {"csv": {"path": tmp_path / "test_read_csv.csv"}}
    output = reader.fit_transform(input)

    assert isinstance(output, dict)
    assert "csv" in output
    assert isinstance(output["csv"], dict)
    assert "path" in output["csv"]
    assert "data" in output["csv"]

    read_df = output["csv"]["data"][["col1", "col2"]]
    assert_frame_equal(df, read_df)

    df.to_csv(tmp_path / "test_read_csv.csv", sep=";")
    input = {"csv": {"path": tmp_path / "test_read_csv.csv"}}
    params = {"csv": {"sep": ";"}}
    output = reader.fit_transform(input, params)

    assert isinstance(output, dict)
    assert "csv" in output
    assert isinstance(output["csv"], dict)
    assert "path" in output["csv"]
    assert "data" in output["csv"]

    read_df = output["csv"]["data"][["col1", "col2"]]
    assert_frame_equal(df, read_df)

    df.to_csv(tmp_path / "test_read_csv.tsv", sep="\t")
    input = {"csv": {"path": tmp_path / "test_read_csv.tsv"}}
    output = reader.fit_transform(input)

    assert isinstance(output, dict)
    assert "csv" in output
    assert isinstance(output["csv"], dict)
    assert "path" in output["csv"]
    assert "data" in output["csv"]

    read_df = output["csv"]["data"][["col1", "col2"]]
    # Check if dataframes are equal
    assert_frame_equal(df, read_df)



"""Provide class for juseless datalad datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from pathlib import Path
from typing import Union

from ....api.decorators import register_datagrabber
from ....datagrabber import PatternDataladDataGrabber


@register_datagrabber
class JuselessDataladUKBVBM(PatternDataladDataGrabber):
    """Juseless UKB VBM Data Grabber class.

    Implements a Data Grabber to access the UKB VBM data in Juseless.

    Parameters
    ----------
    datadir : str or pathlib.Path, optional
        The directory where the datalad dataset will be cloned. If None,
        the datalad dataset will be cloned into a temporary directory
        (default None).

    """

    def __init__(self, datadir: Union[str, Path, None] = None) -> None:
        uri = "ria+http://ukb.ds.inm7.de#~cat_m0wp1"
        rootdir = "m0wp1"
        types = ["VBM_GM"]
        replacements = ["subject", "session"]
        patterns = {"VBM_GM": "m0wp1sub-{subject}_ses-{session}_T1w.nii.gz"}
        super().__init__(
            types=types,
            datadir=datadir,
            uri=uri,
            rootdir=rootdir,
            replacements=replacements,
            patterns=patterns,
        )



"""Provide functions for registry."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from ..utils.logging import logger, raise_error


if TYPE_CHECKING:
    from ..datagrabber import BaseDataGrabber
    from ..storage import BaseFeatureStorage
    from .pipeline_step_mixin import PipelineStepMixin


# Define valid steps for operation
_VALID_STEPS: List[str] = [
    "datagrabber",
    "datareader",
    "preprocessing",
    "marker",
    "storage",
]

# Define registry for valid steps
_REGISTRY: Dict[str, Dict[str, type]] = {x: {} for x in _VALID_STEPS}


def register(step: str, name: str, klass: type) -> None:
    """Register a function to be used in a pipeline step.

    Parameters
    ----------
    step : str
        Name of the step.
    name : str
        Name of the function.
    klass : class
        Class to be registered.

    """
    # Verify step
    if step not in _VALID_STEPS:
        raise_error(msg=f"Invalid step: {step}", klass=ValueError)

    logger.info(f"Registering {name} in {step}")
    _REGISTRY[step][name] = klass


def get_step_names(step: str) -> List[str]:
    """Get the names of the registered functions for a given step.

    Parameters
    ----------
    step : str
        Name of the step.

    Returns
    -------
    list
        List of registered function names.

    """
    # Verify step
    if step not in _VALID_STEPS:
        raise_error(msg=f"Invalid step: {step}", klass=ValueError)

    return list(_REGISTRY[step].keys())


def get_class(step: str, name: str) -> type:
    """Get the class of the registered function for a given step.

    Parameters
    ----------
    step : str
        Name of the step.
    name : str
        Name of the function.

    Returns
    -------
    class
        Registered function class.

    """
    # Verify step
    if step not in _VALID_STEPS:
        raise_error(msg=f"Invalid step: {step}", klass=ValueError)
    # Verify step name
    if name not in _REGISTRY[step]:
        raise_error(msg=f"Invalid name: {name}", klass=ValueError)

    return _REGISTRY[step][name]


def build(
    step: str,
    name: str,
    baseclass: type,
    init_params: Optional[Dict] = None,
) -> Union["BaseDataGrabber", "PipelineStepMixin", "BaseFeatureStorage"]:
    """Ensure that the given object is an instance of the given class.

    Parameters
    ----------
    step : str
        Name of the step.
    name : str
        Name of the function.
    baseclass : class
        Class to be checked against.
    init_parms : dict, optional
        Parameters to pass to the base class constructor (default None).

    Returns
    -------
    object
        An instance of the given base class.

    Raises
    ------
    ValueError
        If the created object with the given name is not an instance of the
        base class.

    """
    # Set default init parameters
    if init_params is None:
        init_params = {}
    # Get class of the registered function
    klass = get_class(step=step, name=name)
    # Create instance of the class
    object_ = klass(**init_params)
    # Verify created instance belongs to the base class
    if not isinstance(object_, baseclass):
        raise_error(
            msg=(
                f"Invalid {step} ({object_.__class__.__name__}). "
                f"Must inherit from {baseclass.__name__}"
            ),
            klass=ValueError,
        )
    return object_



"""Provide concrete implementation for pattern-based datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import re
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np

from ..api.decorators import register_datagrabber
from ..utils import logger, raise_error
from .base import BaseDataGrabber
from .utils import validate_patterns, validate_replacements


@register_datagrabber
class PatternDataGrabber(BaseDataGrabber):
    """Concrete implementation for data grabbing using patterns.

    Implements a Data Grabber that understands patterns to grab data.

    Parameters
    ----------
    types : list of str
        The types of data to be grabbed (default None).
    patterns : dict
        Patterns for each type of data as a dictionary. The keys are the types
        and the values are the patterns. Each occurrence of the string
        `{subject}` in the pattern will be replaced by the indexed element.
    replacements : list of str
        Replacements in the patterns for each item in the "element" tuple.
    datadir : str or pathlib.Path
        The directory where the data is / will be stored.
    """

    def __init__(
        self,
        types: List[str],
        patterns: Dict[str, str],
        replacements: Union[List[str], str],
        datadir: Union[str, Path],
    ) -> None:
        # Validate patterns
        validate_patterns(types=types, patterns=patterns)

        if not isinstance(replacements, list):
            replacements = [replacements]
        # Validate replacements
        validate_replacements(replacements=replacements, patterns=patterns)

        super().__init__(types=types, datadir=datadir)
        logger.debug("Initializing PatternDataGrabber")
        logger.debug(f"\tpatterns = {patterns}")
        logger.debug(f"\treplacements = {replacements}")
        self.patterns = patterns
        self.replacements = replacements

    @property
    def skip_file_check(self) -> bool:
        """Skip file check existence."""
        return False

    def _replace_patterns_regex(
        self, pattern: str
    ) -> Tuple[str, str, List[str]]:
        """Replace the patterns in `pattern` with the named groups.

        It allows elements to be obtained from the filesystem.

        Parameters
        ----------
        pattern : str
            The pattern to be replaced.

        Returns
        -------
        re_pattern : str
            The regular expression with the named groups.
        glob_pattern : str
            The search pattern to be used with glob.
        replacements : list of str
            The replacements present in the pattern.

        """
        re_pattern = pattern
        glob_pattern = pattern
        t_replacements = [
            x for x in self.replacements if f"{{{x}}}" in pattern
        ]

        for t_r in t_replacements:
            # Replace the first of each with a named group definition
            re_pattern = re_pattern.replace(f"{{{t_r}}}", f"(?P<{t_r}>.*)", 1)

        for t_r in t_replacements:
            # Replace the second appearance of each with the named group
            # back reference
            re_pattern = re_pattern.replace(f"{{{t_r}}}", f"(?P={t_r})")

        for t_r in t_replacements:
            glob_pattern = glob_pattern.replace(f"{{{t_r}}}", "*")
        return re_pattern, glob_pattern, t_replacements

    def _replace_patterns_glob(self, element: Dict, pattern: str) -> str:
        """Replace patterns with the element so it can be globbed.

        Parameters
        ----------
        element : dict
            The element to be used in the replacement.
        pattern : str
            The pattern to be replaced.

        Returns
        -------
        str
            The pattern with the element replaced.

        """
        if list(element.keys()) != self.replacements:
            raise_error(
                f"The element keys must be {self.replacements}, "
                f"element has {list(element.keys())}."
            )
        return pattern.format(**element)

    def get_element_keys(self) -> List[str]:
        """Get element keys.

        For each item in the "element" tuple, this functions returns the
        corresponding key, that is, the ``replacements`` of patterns defined
        in the constructor.

        Returns
        -------
        list of str
            The element keys.

        """
        return self.replacements

    def get_item(self, **element: str) -> Dict[str, Dict]:
        """Implement single element indexing in the database.

        This method constructs a real path to the requested item's data, by
        replacing the ``patterns`` with actual values passed via ``**element``.

        Parameters
        ----------
        element : dict
            The element to be indexed. The keys must be the same as the
            replacements.

        Returns
        -------
        dict
            Dictionary of dictionaries for each type of data required for the
            specified element.

        """
        out = {}
        for t_type in self.types:
            t_pattern = self.patterns[t_type]
            t_replace = self._replace_patterns_glob(element, t_pattern)
            if "*" in t_replace:
                t_matches = list(self.datadir.absolute().glob(t_replace))
                if len(t_matches) > 1:
                    raise_error(
                        f"More than one file matches for {element} / {t_type}:"
                        f" {t_matches}"
                    )
                elif len(t_matches) == 0:
                    raise_error(f"No file matches for {element} / {t_type}")
                t_out = t_matches[0]
            else:
                t_out = self.datadir / t_replace
                if not self.skip_file_check:
                    if not t_out.exists() and not t_out.is_symlink():
                        raise_error(
                            f"Cannot access {t_type} for {element}: "
                            f"File {t_out} does not exist"
                        )
            out[t_type] = {"path": t_out}
        return out

    def get_elements(self) -> List:
        """Implement fetching list of elements in the dataset.

        It will use regex to search for "replacements" in the "patterns" and
        return the intersection of the results for each type i.e., build a
        list of elements that have all the required types.

        Returns
        -------
        list
            The list of elements that can be grabbed in the dataset.
        """
        elements = None

        # Iterate by number of replacements. At least one pattern must have
        # all of them.
        replacement_in_patterns = [
            np.sum([x in self.patterns[t_type] for x in self.replacements])
            for t_type in self.types
        ]
        order = np.argsort(replacement_in_patterns)

        # Start from the one with the most replacements and then
        # filter out elements based on the ones with less replacements.
        for t_idx in reversed(order):
            t_type = self.types[t_idx]
            types_element = set()
            # Get the pattern
            t_pattern = self.patterns[t_type]
            # Replace the pattern
            (
                re_pattern,
                glob_pattern,
                t_replacements,
            ) = self._replace_patterns_regex(t_pattern)
            for fname in self.datadir.glob(glob_pattern):
                suffix = fname.relative_to(self.datadir).as_posix()
                m = re.match(re_pattern, suffix)
                if m is not None:
                    # Find the groups of replacements present in the pattern
                    # If one replacement is not present, set it to None.
                    # We will take care of this in the intersection
                    t_element = tuple([m.group(k) for k in t_replacements])
                    if len(self.replacements) == 1:
                        t_element = t_element[0]
                    types_element.add(t_element)
            # TODO: does this make sense as elements is always None
            if elements is None:
                elements = types_element
            else:
                # Do the intersection by filtering out elements in which
                # the replacements are not None
                if t_replacements == self.replacements:
                    elements.intersection(types_element)
                else:
                    t_repl_idx = [
                        i
                        for i, v in enumerate(self.replacements)
                        if v in t_replacements
                    ]
                    new_elements = set()
                    for t_element in elements:
                        if (
                            tuple(np.array(t_element)[t_repl_idx])
                            in types_element
                        ):
                            new_elements.add(t_element)
                    elements = new_elements
        if elements is None:
            elements = set()
        return list(elements)



"""Provide tests for data utils."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

from typing import List

import numpy as np
import pytest

from junifer.data.utils import closest_resolution


@pytest.mark.parametrize(
    "resolution, valid_resolutions, expected",
    [
        (1.0, [1.0, 2.0, 3.0], 1.0),
        (1.1, [1.0, 2.0, 3.0], 1.0),
        (0.9, [1.0, 2.0, 3.0], 1.0),
        (2.1, [1.0, 2.0, 3.0], 2.0),
        (2.0, [1.0, 2.0, 3.0], 2.0),
        (4.0, [1.0, 2.0, 3.0], 3.0),
        (None, [1.0, 2.0, 3.0], 1.0),
    ],
)
def test_closest_resolution(
    resolution: float, valid_resolutions: List[float], expected: float
):
    """Test closest_resolution.

    Parameters
    ----------
    resolution: float
        The resolution to test.
    valid_resolutions: list of float
        The valid resolutions.
    expected: float
        The expected result.
    """
    assert closest_resolution(resolution, valid_resolutions) == expected
    assert (
        closest_resolution(resolution, np.array(valid_resolutions)) == expected
    )



"""Provide tests for juseless datagrabber."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Leonard Sasse <l.sasse@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import socket

import pytest

from junifer.datagrabber.hcp import DataladHCP1200
from junifer.utils.logging import configure_logging


# Check if the test is running on juseless
if socket.gethostname() != "juseless":
    pytest.skip("These tests are only for juseless", allow_module_level=True)

configure_logging(level="DEBUG")


def test_juselessdataladhcp_datagrabber() -> None:
    """Test datalad HCP datagrabber."""
    with DataladHCP1200() as dg:
        all_elements = dg.get_elements()
        test_element = all_elements[0]

        out = dg[test_element]

        assert out["BOLD"]["path"].exists()
        assert out["BOLD"]["path"].isfile()



"""Provide functions for parser."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

import importlib
import importlib.util
import os
import sys
from pathlib import Path
from typing import Dict, Union

import yaml

from ..utils.logging import logger, raise_error


def parse_yaml(filepath: Union[str, Path]) -> Dict:
    """Parse YAML.

    Parameters
    ----------
    filepath : str or pathlib.Path
        The filepath to read from.

    Returns
    -------
    dict
        The contents represented as dictionary.

    """
    # Convert str to Path
    if not isinstance(filepath, Path):
        filepath = Path(filepath)

    logger.info(f"Parsing yaml file: {str(filepath.absolute())}")
    # Filepath existence check
    if not filepath.exists():
        raise_error(f"File does not exist: {str(filepath.absolute())}")
    # Filepath reading
    with open(filepath, "r") as f:
        contents = yaml.safe_load(f)
    if "elements" in contents:
        if contents["elements"] is None:
            raise_error(
                "The elements key was defined but its content is empty. "
                "Please define the elements to operate on or remove the key."
            )
    # load modules
    if "with" in contents:
        to_load = contents["with"]
        # Convert load modules to list
        if not isinstance(to_load, list):
            to_load = [to_load]
        for t_module in to_load:
            if t_module.endswith(".py"):
                logger.debug(f"Importing file: {t_module}")
                file_path = Path(os.getcwd()) / t_module
                if not file_path.exists():
                    raise_error(
                        f"File in 'with' section does not exist: {file_path}"
                    )
                spec = importlib.util.spec_from_file_location(
                    t_module, file_path
                )
                module = importlib.util.module_from_spec(spec)  # type: ignore
                sys.modules[t_module] = module
                spec.loader.exec_module(module)  # type: ignore
            else:
                logger.info(f"Importing module: {t_module}")
                importlib.import_module(t_module)

    return contents



"""Provide functions for list of coordinates."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
# License: AGPL

import typing
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike

from ..utils.logging import logger, raise_error


# Path to the VOIs
_vois_path = Path(__file__).parent / "VOIs"

# Path to the metadata of the VOIs
_vois_meta_path = _vois_path / "meta"

"""
A dictionary containing all supported coordinates and their respective file or
data.

The built-in coordinates are files that are shipped with the package in the
data/VOIs directory. The user can also register their own coordinates, which
will be stored as numpy arrays in the dictionary.
"""
_available_coordinates: Dict[
    str, Union[Path, Dict[str, Union[ArrayLike, List[str]]]]
] = {
    "CogAC": _vois_meta_path / "CogAC_VOIs.txt",
    "CogAR": _vois_meta_path / "CogAR_VOIs.txt",
    "DMNBuckner": _vois_meta_path / "DMNBuckner_VOIs.txt",
    "eMDN": _vois_meta_path / "eMDN_VOIs.txt",
    "Empathy": _vois_meta_path / "Empathy_VOIs.txt",
    "eSAD": _vois_meta_path / "eSAD_VOIs.txt",
    "extDMN": _vois_meta_path / "extDMN_VOIs.txt",
    "Motor": _vois_meta_path / "Motor_VOIs.txt",
    "MultiTask": _vois_meta_path / "MultiTask_VOIs.txt",
    "PhysioStress": _vois_meta_path / "PhysioStress_VOIs.txt",
    "Rew": _vois_meta_path / "Rew_VOIs.txt",
    "Somatosensory": _vois_meta_path / "Somatosensory_VOIs.txt",
    "ToM": _vois_meta_path / "ToM_VOIs.txt",
    "VigAtt": _vois_meta_path / "VigAtt_VOIs.txt",
    "WM": _vois_meta_path / "WM_VOIs.txt",
}


def register_coordinates(
    name: str,
    coordinates: ArrayLike,
    voi_names: List[str],
    overwrite: Optional[bool] = False,
) -> None:
    """Register coordinates.

    Parameters
    ----------
    name : str
        The name of the coordinates.
    coordinates : numpy.ndarray
        The coordinates.
    voi_names : list of str
        The names of the VOIs.
    overwrite : bool, optional
        If True, overwrite an existing list of coordinates with the same name.
        Does not apply to built-in coordinates (default False).
    """
    if name in _available_coordinates:
        if isinstance(_available_coordinates[name], Path):
            raise_error(
                f"Coordinates {name} already registered as built-in "
                "coordinates."
            )
        if overwrite is True:
            logger.info(f"Overwriting coordinates {name}")
        else:
            raise_error(
                f"Coordinates {name} already registered. "
                "Set `overwrite=True` to update its value."
            )

    if not isinstance(coordinates, np.ndarray):
        raise_error(
            f"Coordinates must be a numpy.ndarray, not {type(coordinates)}."
        )
    if coordinates.ndim != 2:
        raise_error(
            f"Coordinates must be a 2D array, not {coordinates.ndim}D."
        )
    if coordinates.shape[1] != 3:
        raise_error(
            f"Each coordinate must have 3 values, not {coordinates.shape[1]} "
        )
    if len(voi_names) != coordinates.shape[0]:
        raise_error(
            f"Length of voi_names ({len(voi_names)}) does not match the "
            f"number of coordinates ({coordinates.shape[0]})."
        )
    _available_coordinates[name] = {
        "coords": coordinates,
        "voi_names": voi_names,
    }


def list_coordinates() -> List[str]:
    """List all the available coordinates lists (VOIs).

    Returns
    -------
    list of str
        A list with all available coordinates names.
    """
    return sorted(_available_coordinates.keys())


def load_coordinates(name: str) -> Tuple[ArrayLike, List[str]]:
    """Load coordinates.

    Parameters
    ----------
    name : str
        The name of the coordinates.

    Returns
    -------
    numpy.ndarray
        The coordinates.
    list of str
        The names of the VOIs.
    """
    if name not in _available_coordinates:
        raise_error(f"Coordinates {name} not found.")
    t_coord = _available_coordinates[name]
    if isinstance(t_coord, Path):
        df_coords = pd.read_csv(t_coord, sep="\t", header=None)
        coords = df_coords.iloc[:, [0, 1, 2]].to_numpy()
        names = [x for x in df_coords.iloc[:, [3]].values[:, 0]]
    else:
        coords = t_coord["coords"]
        coords = typing.cast(ArrayLike, coords)
        names = t_coord["voi_names"]
        names = typing.cast(List[str], names)
    return coords, names



"""Provide abstract base class for markers."""

# Authors: Federico Raimondo <f.raimondo@fz-juelich.de>
#          Synchon Mandal <s.mandal@fz-juelich.de>
# License: AGPL

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from ..pipeline import PipelineStepMixin, UpdateMetaMixin
from ..utils import logger, raise_error


if TYPE_CHECKING:
    from junifer.storage import BaseFeatureStorage


class BaseMarker(ABC, PipelineStepMixin, UpdateMetaMixin):
    """Abstract base class for all markers.

    Parameters
    ----------
    on : str or list of str
        The kind of data to apply the marker to. By default, will work on all
        available data.
    name : str, optional
        The name of the marker. By default, it will use the class name as the
        name of the marker (default None).

    """

    def __init__(
        self,
        on: Optional[Union[List[str], str]] = None,
        name: Optional[str] = None,
    ) -> None:
        if on is None:
            on = self.get_valid_inputs()
        if not isinstance(on, list):
            on = [on]
        self.name = self.__class__.__name__ if name is None else name

        if any(x not in self.get_valid_inputs() for x in on):
            wrong_on = [x for x in on if x not in self.get_valid_inputs()]
            raise ValueError(f"{self.name} cannot be computed on {wrong_on}")
        self._on = on

    def get_valid_inputs(self) -> List[str]:
        """Get valid data types for input.

        Returns
        -------
        list of str
            The list of data types that can be used as input for this marker.
        """
        raise_error(
            msg="Concrete classes need to implement get_valid_inputs().",
            klass=NotImplementedError,
        )

    def validate_input(self, input: List[str]) -> None:
        """Validate input.

        Parameters
        ----------
        input : list of str
            The input to the pipeline step. The list must contain the
            available Junifer Data dictionary keys.

        Raises
        ------
        ValueError
            If the input does not have the required data.

        """
        if not any(x in input for x in self._on):
            raise_error(
                "Input does not have the required data."
                f"\t Input: {input}"
                f"\t Required (any of): {self._on}"
            )

    @abstractmethod
    def get_output_type(self, input_type: str) -> str:
        """Get output type.

        Parameters
        ----------
        input_type : str
            The data type input to the marker.

        Returns
        -------
        str
            The storage type output by the marker.

        """
        raise_error(
            msg="Concrete classes need to implement get_output_type().",
            klass=NotImplementedError,
        )

    @abstractmethod
    def compute(self, input: Dict, extra_input: Optional[Dict] = None) -> Dict:
        """Compute.

        Parameters
        ----------
        input : dict
            A single input from the pipeline data object in which to compute
            the marker.
        extra_input : dict, optional
            The other fields in the pipeline data object. Useful for accessing
            other data kind that needs to be used in the computation. For
            example, the functional connectivity markers can make use of the
            confounds if available (default None).

        Returns
        -------
        dict
            The computed result as dictionary. This will be either returned
            to the user or stored in the storage by calling the store method
            with this as a parameter.

        """
        raise_error(
            msg="Concrete classes need to implement compute().",
            klass=NotImplementedError,
        )

    def store(
        self,
        type_: str,
        out: Dict[str, Any],
        storage: "BaseFeatureStorage",
    ) -> None:
        """Store.

        Parameters
        ----------
        type_ : str
            The data type to store.
        out : dict
            The computed result as a dictionary to store.
        storage : storage-like
            The storage class, for example, SQLiteFeatureStorage.

        """
        output_type_ = self.get_output_type(type_)
        logger.debug(f"Storing {output_type_} in {storage}")
        storage.store(kind=output_type_, **out)

    def fit_transform(
        self,
        input: Dict[str, Dict],
        storage: Optional["BaseFeatureStorage"] = None,
    ) -> Dict:
        """Fit and transform.

        Parameters
        ----------
        input : dict
            The Junifer Data object.
        storage : storage-like, optional
            The storage class, for example, SQLiteFeatureStorage.

        Returns
        -------
        dict
            The processed output as a dictionary. If `storage` is provided,
            empty dictionary is returned.

        """
        out = {}
        for type_ in self._on:
            if type_ in input.keys():
                logger.info(f"Computing {type_}")
                t_input = input[type_]
                extra_input = input.copy()
                extra_input.pop(type_)
                t_meta = t_input["meta"].copy()
                t_meta["type"] = type_

                t_out = self.compute(input=t_input, extra_input=extra_input)
                t_out["meta"] = t_meta

                self.update_meta(t_out, "marker")

                if storage is not None:
                    logger.info(f"Storing in {storage}")
                    self.store(type_=type_, out=t_out, storage=storage)
                else:
                    logger.info("No storage specified, returning dictionary")
                    out[type_] = t_out

        return out
